from ..utils.utils import get_logger

logger = get_logger(__name__)

ENV = {
    'path': '.env',
    'content': """FAB_CLIENT_ID=your_client_id_here
FAB_CLIENT_SECRET=your_client_secret_here
FAB_TENANT_ID=your_tenant_id_here
FAB_USERNAME=your_username_here
FAB_PASSWORD=your_password_here
DATABASE_USERNAME=your_database_username_here
DATABASE_PASSWORD=your_database_password_here
GH_TOKEN=your_github_token_here""",
}


BRANCHES = {
    'path': 'branches.json',
    'content': """{
    "main": "-PRD",
    "master": "-PRD",
    "dev": "-DEV",
    "develop": "-DEV",
    "staging": "-STG"
}""",
}


WORKSPACES_ROLES = {
    'path': 'workspaces_roles.json',
    'content': """[
    {
        "user_uuid": "00000000-0000-0000-0000-0000000000000",
        "user_type": "User",
        "role": "Admin"
    },
    {
        "user_uuid": "00000000-0000-0000-0000-0000000000000",
        "user_type": "Group",
        "role": "Member"
    },
    {
        "user_uuid": "00000000-0000-0000-0000-0000000000000",
        "user_type": "ServicePrincipal",
        "role": "Contributor"
    },
    {
        "user_uuid": "00000000-0000-0000-0000-0000000000000",
        "user_type": "ServicePrincipalProfile",
        "role": "Viewer"
    }
]""",
}


CONNECTIONS_ROLES = {
    'path': 'connections_roles.json',
    'content': """[
        {
            "user_uuid": "00000000-0000-0000-0000-0000000000000",
            "user_type": "User",
            "role": "Owner"
        },
        {
            "user_uuid": "00000000-0000-0000-0000-0000000000000",
            "user_type": "Group",
            "role": "User"
        },
        {
            "user_uuid": "00000000-0000-0000-0000-0000000000000",
            "user_type": "ServicePrincipal",
            "role": "UserWithReshare"
        },
        {
            "user_uuid": "00000000-0000-0000-0000-0000000000000",
            "user_type": "ServicePrincipalProfile",
            "role": "UserWithReshare"
        }
    ]""",
}


GITIGNORE = {
    'path': '.gitignore',
    'content': """**/.pbi/localSettings.json
**/.pbi/cache.abf
**/__pycache__/**
**/_stg/**
.vscode/
.venv
.env
**/py_fab.egg-info
**/dist
**/build
metadata/""",
}


GITATTRIBUTES = {
    'path': '.gitattributes',
    'content': """src/**/config.json merge=union
# This file is used to define attributes for paths in the repository.
# The 'merge=union' attribute allows for union merging of JSON files in the 'src' directory.
# This means that when merging changes, if there are conflicts, the resulting file will contain all unique elements from the conflicting files.
""",
}


SRC = {
    'path': 'src/README.md',
    'content': """(# Source Directory
  This directory contains the source code for the project. 
  It is structured to facilitate development and deployment of the application.""",
}

import os


def create_support_files():
    """
    Create support files with predefined content for PyFabricOps CI/CD operations.
    """
    files = [
        ENV,
        BRANCHES,
        WORKSPACES_ROLES,
        CONNECTIONS_ROLES,
        GITIGNORE,
        GITATTRIBUTES,
        SRC,
    ]

    # Create directories and files
    for file_dict in files:
        path = file_dict.get('path')
        if path:
            os.makedirs(os.path.dirname(path) or '.', exist_ok=True)
            with open(path, 'w', encoding='utf-8') as f:
                f.write(file_dict['content'])
            logger.success(f'Created {path}')



from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_folders(
    workspace: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    List folders in a workspace

    Args:
        workspace (str): The workspace to list folders from.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of folders in the workspace.

    Examples:
        ```python
        list_folders('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    return api_request(
        '/workspaces/' + resolve_workspace(workspace) + '/folders',
        support_pagination=True,
    )


def get_folder_id(workspace: str, folder_name: str) -> Union[str, None]:
    """
    Retrieves the ID of a folder by its name.

    Args:
        folder_name (str): The name of the folder.

    Returns:
        str | None: The ID of the folder if found, otherwise None.
    """
    folders = list_folders(
        workspace=resolve_workspace(workspace),
        df=False,
    )
    for _folder in folders:
        if _folder['displayName'] == folder_name:
            return _folder['id']
    logger.warning(
        f"Folder '{folder_name}' not found in workspace '{workspace}'."
    )
    return None


def resolve_folder(workspace: str, folder: str) -> Union[str, None]:
    """
    Resolves a folder name to its ID.

    Args:
        workspace (str): The name or ID of the workspace.
        folder (str): The name or ID of the folder.

    Returns:
        str | None: The ID of the folder if found, otherwise None.
    """
    if is_valid_uuid(folder):
        return folder
    else:
        return get_folder_id(workspace, folder)


@df
def get_folder(
    workspace: str, folder: str, *, df: Optional[bool] = True
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Get a folder in a workspace.

    Args:
        workspace (str): The name or id of the workspace to get the folder from.
        folder (str): The name or id of the folder to get.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The folder details if found, otherwise None.

    Examples:
        ```python
        get_folder(
            workspace_id='123e4567-e89b-12d3-a456-426614174000',
            folder_id='98f6b7c8-1234-5678-90ab-cdef12345678'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)
    return api_request(
        '/workspaces/'
        + workspace_id
        + '/folders/'
        + resolve_folder(workspace_id, folder),
    )


@df
def create_folder(
    workspace: str,
    display_name: str,
    *,
    parent_folder: str = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Create a new folder in the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace where the folder will be created.
        display_name (str): The name of the folder to create.
        parent_folder (str): The name or ID of the parent folder.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created folder details if successful, otherwise None.

    Examples:
        ```python
        create_folder(
            workspace_id='123e4567-e89b-12d3-a456-426614174000',
            display_name='NewFolder',
            parent_folder_id='456e7890-e12b-34d5-a678-90abcdef1234'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {'displayName': display_name}

    if parent_folder:
        payload['parentFolderId'] = resolve_folder(workspace_id, parent_folder)

    return api_request(
        '/workspaces/' + workspace_id + '/folders',
        payload=payload,
        method='post',
    )


def delete_folder(workspace: str, folder: str) -> None:
    """
    Delete a folder in a workspace

    Args:
        workspace (str): The name or ID of the workspace to delete the folder from.
        folder (str): The name or ID of the folder to delete.

    Returns:
        None.

    Examples:
        ```python
        delete_folder(
            workspace_id='123e4567-e89b-12d3-a456-426614174000',
            folder_id='98f6b7c8-1234-5678-90ab-cdef12345678'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    return api_request(
        '/workspaces/'
        + workspace_id
        + '/folders/'
        + resolve_folder(workspace_id, folder),
        method='delete',
    )


@df
def update_folder(
    workspace: str,
    folder: str,
    display_name: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Update a existing folder in the specified workspace.

    Args:
        workspace (str): The name or id of the workspace where the folder will be updated.
        folder (str): The name or id of the folder to update.
        display_name (str): The name of the folder to update.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated folder details if successful, otherwise None.

    Examples:
        ```python
        update_folder(
            workspace_id='123e4567-e89b-12d3-a456-426614174000',
            folder_id='98f6b7c8-1234-5678-90ab-cdef12345678',
            display_name='NewFolderName',
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {'displayName': display_name}

    return api_request(
        '/workspaces/'
        + workspace_id
        + '/folders/'
        + resolve_folder(workspace_id, folder),
        payload=payload,
        method='patch',
    )


@df
def move_folder(
    workspace: str,
    folder: str,
    *,
    target_folder: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Move a existing folder into other or root folder.

    Args:
        workspace (str): The workspace where the folder will be updated.
        folder (str): The folder to be moved.
        target_folder (str): The name of the parent folder will receive the moved folder.
        df (bool, optional): Keyword-only.
            If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The moved folder details if successful, otherwise None.

    Examples:
        ```python
        move_folder(
            workspace_id='123e4567-e89b-12d3-a456-426614174000',
            folder_id='414e7890-e12b-34d5-a678-90abcdef1234',
            target_folder_id='9859b7c8-1234-5678-90ab-cdef12345678',
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {}

    if target_folder:
        payload = {
            'targetFolderId': resolve_folder(workspace_id, target_folder)
        }

    return api_request(
        '/workspaces/'
        + workspace_id
        + '/folders/'
        + resolve_folder(workspace_id, folder)
        + '/move',
        payload=payload,
        method='post',
    )



from typing import Any, Dict, List, Literal, Optional, Union
from venv import create

from pandas import DataFrame

from pyfabricops.utils.exceptions import OptionNotAvailableError

from ..api.api import api_request
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_deployment_pipelines(
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    List deployment pipelines

    Args:
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of deployment pipelines.

    Examples:
        ```python
        list_deployment_pipelines()
        ```
    """
    return api_request(
        '/deploymentPipelines',
        support_pagination=True,
    )


def get_deployment_pipeline_id(pipeline_name: str) -> Union[str, None]:
    """
    Retrieves the ID of a deployment pipeline by its name.

    Args:
        pipeline_name (str): The name of the deployment pipeline.

    Returns:
        str | None: The ID of the deployment pipeline if found, otherwise None.
    """
    pipelines = list_deployment_pipelines(df=False)
    for _pipeline in pipelines:
        if _pipeline['displayName'] == pipeline_name:
            return _pipeline['id']

    logger.warning(f"Deployment pipeline '{pipeline_name}' not found.")
    return None


def resolve_deployment_pipeline(pipeline: str) -> Union[str, None]:
    """
    Resolves a deployment pipeline to its ID.

    Args:
        pipeline (str): The name or id of the deployment pipeline.

    Returns:
        str | None: The ID of the deployment pipeline if found, otherwise None.
    """
    if is_valid_uuid(pipeline):
        return pipeline
    else:
        return get_deployment_pipeline_id(pipeline)


@df
def get_deployment_pipeline(
    pipeline: str, df: Optional[bool] = True
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns the specified deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline to retrieve.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
                If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]) The details of the deployment pipeline if found, otherwise None. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        get_deployment_pipeline('123e4567-e89b-12d3-a456-426614174000')
        get_deployment_pipeline('MyProjectPipeline')
        get_deployment_pipeline('MyProjectPipeline', df=False) # Returns as list
        ```
    """
    return api_request(
        '/deploymentPipelines/' + resolve_deployment_pipeline(pipeline)
    )


def resolve_deployment_pipeline_stage(
    pipeline: str, stage_name: str
) -> Union[str, None]:
    """
    Resolves a deployment pipeline stage to its ID.

    Args:
        pipeline (str): The name or id of the deployment pipeline.
        stage_name (str): The name of the deployment pipeline stage.

    Returns:
        str | None: The ID of the deployment pipeline stage if found, otherwise None.
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None
    pipeline_details = get_deployment_pipeline(pipeline, df=False)
    for stage in pipeline_details.get('stages', []):
        if stage['displayName'] == stage_name:
            return stage['id']

    logger.warning(
        f"Stage '{stage_name}' not found in deployment pipeline '{pipeline}'."
    )
    return None


@df
def create_deployment_pipeline(
    display_name: str,
    stages: List[Dict[str, Any]],
    *,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Creates a new deployment pipeline.

    Args:
        display_name (str): The display name of the deployment pipeline.
        stages (List[Dict[str, Any]]): A list of stages for the deployment pipeline.
        description (Optional[str]): An optional description for the deployment pipeline.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): The details of the created deployment pipeline.

    Examples:
        ```python
        create_deployment_pipeline(
            display_name='My Deployment Pipeline',
            stages=[
                {
                    "displayName": "Development",
                    "description": "Development stage description",
                    "isPublic": false
                },
                {
                    "displayName": "Test",
                    "description": "Test stage description",
                    "isPublic": false
                },
                {
                    "displayName": "Production",
                    "description": "Production stage description",
                    "isPublic": true
                }
            ],
            description='This is my deployment pipeline',
            df=True
        )
        ```
    """
    payload = {
        'displayName': display_name,
        'stages': stages,
    }
    if description:
        payload['description'] = description
    return api_request(
        '/deploymentPipelines',
        method='POST',
        payload=payload,
    )


def assign_workspace_to_stage(
    workspace: str,
    pipeline: str,
    stage: str,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Assigns a workspace to a deployment pipeline stage.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        stage (str): The name or ID of the stage within the deployment pipeline.
        workspace (str): The name or ID of the workspace to assign.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): The updated details of the deployment pipeline stage.

    Examples:
        ```python
        assign_workspace_to_stage(
            workspace='My Workspace',
            pipeline='My Deployment Pipeline',
            stage='Development'
        )
        ```
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None
    stage_id = resolve_deployment_pipeline_stage(pipeline, stage)
    if not stage_id:
        return None
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    return api_request(
        f'/deploymentPipelines/{pipeline_id}/stages/{stage_id}/assignWorkspace',
        method='POST',
        payload={'workspaceId': workspace_id},
    )


def unassign_workspace_to_stage(
    pipeline: str,
    stage: str,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Unassigns a workspace from a deployment pipeline stage.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        stage (str): The name or ID of the stage within the deployment pipeline.
        workspace (str): The name or ID of the workspace to unassign.
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None
    stage_id = resolve_deployment_pipeline_stage(pipeline, stage)
    if not stage_id:
        return None

    return api_request(
        f'/deploymentPipelines/{pipeline_id}/stages/{stage_id}/unassignWorkspace',
        method='POST',
    )


@df
def add_deployment_pipeline_role_assignment(
    pipeline: str,
    user_uuid: str,
    user_type: Literal[
        'User', 'Group', 'ServicePrincipal', 'ServicePrincipalProfile'
    ] = 'User',
    role: Literal['Admin', 'Contributor', 'Member', 'Viewer'] = 'Admin',
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Adds a permission to a deployment pipeline for a user.

    Args:
        pipeline (str): The ID or name of the deployment pipeline.
        user_uuid (str): The UUID of the user.
        user_type (str): The type of user (options: User, Group, ServicePrincipal, ServicePrincipalProfile).
        role (str): The role to assign (options: admin, member, contributor, viewer).
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The role assignment details if successful.

    Raises:
        ResourceNotFoundError: If the specified deployment pipeline is not found.
        OptionNotAvailableError: If the user type or role is invalid.

    Examples:
        ```python
        add_deployment_pipeline_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'FefEFewf-feF-1234-5678-9abcdef01234', user_type='User', role='Admin'
        )
        ```
    """
    if user_type not in [
        'User',
        'Group',
        'ServicePrincipal',
        'ServicePrincipalProfile',
    ]:
        raise OptionNotAvailableError(
            f'Invalid user type: {user_type}. Must be one of: User, Group, ServicePrincipal, ServicePrincipalProfile'
        )
    if role not in ['Admin', 'Contributor', 'Member', 'Viewer']:
        raise OptionNotAvailableError(
            f'Invalid role: {role}. Must be one of: Admin, Contributor, Member, Viewer'
        )
    payload = {'principal': {'id': user_uuid, 'type': user_type}, 'role': role}

    return api_request(
        '/deploymentPipelines/'
        + resolve_deployment_pipeline(pipeline)
        + '/roleAssignments',
        payload=payload,
        method='post',
    )


@df
def list_deployment_pipeline_role_assignments(
    pipeline: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Lists all role assignments for a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline to list role assignments for.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of role assignments.

    Examples:
        ```python
        list_deployment_pipeline_role_assignments('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    return api_request(
        '/deploymentPipelines/'
        + resolve_deployment_pipeline(pipeline)
        + '/roleAssignments',
        method='GET',
        support_pagination=True,
    )


@df
def deploy_stage_content(
    pipeline: str,
    source_stage: str,
    target_stage: str,
    *,
    items: Optional[List[Dict[str, Any]]] = None,
    note: Optional[str] = None,
    options: Optional[Dict[str, Any]] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Deploys content to a specified stage in a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        source_stage (str): The name or ID of the source stage within the deployment pipeline.
        target_stage (str): The name or ID of the target stage within the deployment pipeline.
        items (Optional[List[str]]): A list of item IDs to deploy. If not provided, all items will be deployed.
        note (Optional[str]): An optional note for the deployment.
        options (Optional[Dict[str, Any]]): Additional deployment options.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): The deployment result.

    Examples:
        ```python
        deploy_stage_content(
            '123e4567-e89b-12d3-a456-426614174000',
            'develop',
            'staging',
        )

        deploy_stage_content(
            pipeline = '123e4567-e89b-12d3-a456-426614174000',
            source_stage = 'develop',
            target_stage = 'staging',
            items = [
                {
                "sourceItemId": "1a201f2a-d1d8-45c0-8c61-1676338517de",
                "itemType": "SemanticModel"
                },
                {
                "sourceItemId": "2d225191-65f8-4ec3-b77d-06100602b1f7",
                "itemType": "Report"
                }
            ],
            note = "Deploying selected items from develop to staging",
            options = {"allowCrossRegionDeployment": True}
        )
        ```
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None
    source_stage_id = resolve_deployment_pipeline_stage(
        pipeline_id, source_stage
    )
    if not source_stage_id:
        return None
    target_stage_id = resolve_deployment_pipeline_stage(
        pipeline_id, target_stage
    )
    if not target_stage_id:
        return None

    payload = {
        'sourceStageId': source_stage_id,
        'targetStageId': target_stage_id,
    }

    if items:
        payload['items'] = items
    if note:
        payload['note'] = note
    if options:
        payload['options'] = options

    return api_request(
        '/deploymentPipelines/'
        + resolve_deployment_pipeline(pipeline)
        + '/deploy',
        method='POST',
        payload=payload,
        support_lro=True,
    )


@df
def list_deployment_pipeline_operations(
    pipeline: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Lists all operations for a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline to list operations for.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of deployment pipeline operations.
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None

    return api_request(
        '/deploymentPipelines/' + pipeline_id + '/operations',
        method='GET',
        support_pagination=True,
    )


@df
def get_deployment_pipeline_operation(
    pipeline: str,
    operation_id: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Gets a specific operation for a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        operation_id (str): The ID of the operation to retrieve.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The requested deployment pipeline operation.
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None

    return api_request(
        '/deploymentPipelines/' + pipeline_id + '/operations/' + operation_id,
        method='GET',
    )


@df
def update_deployment_pipeline(
    pipeline: str,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates a deployment pipeline's details.

    Args:
        pipeline (str): The name or ID of the deployment pipeline to update.
        display_name (Optional[str]): The new display name for the deployment pipeline.
        description (Optional[str]): The new description for the deployment pipeline.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated deployment pipeline details.

    Examples:
        ```python
        updated_pipeline = update_deployment_pipeline(
            pipeline="my-pipeline",
            display_name="My Updated Pipeline",
            description="This is an updated description."
        )
        ```
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None

    payload = {}
    if display_name:
        payload['displayName'] = display_name
    if description:
        payload['description'] = description

    return api_request(
        '/deploymentPipelines/' + pipeline_id,
        method='PATCH',
        payload=payload,
    )


def update_deployment_pipeline_stage(
    pipeline: str,
    stage: str,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    is_public: Optional[bool] = None,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates a deployment pipeline stage's details.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        stage (str): The name or ID of the stage within the deployment pipeline.
        display_name (Optional[str]): The new display name for the stage.
        description (Optional[str]): The new description for the stage.
        is_public (Optional[bool]): Whether the stage is public or not.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated deployment pipeline stage details.
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None
    stage_id = resolve_deployment_pipeline_stage(pipeline, stage)
    if not stage_id:
        return None

    payload = {}
    if display_name:
        payload['displayName'] = display_name
    if description:
        payload['description'] = description
    if is_public is not None:
        payload['isPublic'] = is_public

    return api_request(
        f'/deploymentPipelines/{pipeline_id}/stages/{stage_id}',
        method='PATCH',
        payload=payload,
    )


def delete_deployment_pipeline(
    pipeline: str,
) -> None:
    """
    Deletes a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline to delete.

    Returns:
        None

    Examples:
        ```python
        delete_deployment_pipeline('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None

    api_request(
        '/deploymentPipelines/' + pipeline_id,
        method='DELETE',
    )


def delete_deployment_pipeline_role_assignment(
    pipeline: str,
    role_assignment_id: str,
) -> None:
    """
    Deletes a role assignment from a deployment pipeline.

    Args:
        pipeline (str): The name or ID of the deployment pipeline.
        role_assignment_id (str): The ID of the role assignment to delete.

    Returns:
        None

    Examples:
        ```python
        delete_deployment_pipeline_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'role-assignment-uuid'
        )
        ```
    """
    pipeline_id = resolve_deployment_pipeline(pipeline)
    if not pipeline_id:
        return None

    api_request(
        '/deploymentPipelines/'
        + pipeline_id
        + '/roleAssignments/'
        + role_assignment_id,
        method='DELETE',
    )



from typing import Any, Dict, List, Literal, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.capacities import resolve_capacity
from ..utils.decorators import df
from ..utils.exceptions import OptionNotAvailableError
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_workspaces(
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns a list of workspaces.

    Args:
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (DataFrame | list | None): A list of workspaces or a DataFrame if df is True.

    Examples:
        ```python
        list_workspaces() # Returns as DataFrame
        list_workspaces(df=False) # Returns as list
        ```
    """
    return api_request('/workspaces', support_pagination=True)


def get_workspace_id(workspace: str) -> Union[str, None]:
    """
    Retrieves the ID of a workspace by its name.

    Args:
        workspace (str): The name of the workspace.

    Returns:
        str | None: The ID of the workspace if found, otherwise None.
    """
    workspaces = list_workspaces(df=False)
    for _workspace in workspaces:
        if _workspace['displayName'] == workspace:
            return _workspace['id']

    logger.warning(f"Workspace '{workspace}' not found.")
    return None


def resolve_workspace(workspace: str) -> Union[str, None]:
    """
    Resolves a workspace name to its ID.

    Args:
        workspace (str): The name of the workspace.

    Returns:
        str | None: The ID of the workspace if found, otherwise None.
    """
    if is_valid_uuid(workspace):
        return workspace
    else:
        return get_workspace_id(workspace)


@df
def get_workspace(
    workspace: str, df: Optional[bool] = True
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace to retrieve.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
                If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]) The details of the workspace if found, otherwise None. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        get_workspace('123e4567-e89b-12d3-a456-426614174000')
        get_workspace('MyProjectWorkspace')
        get_workspace('MyProjectWorkspace', df=False) # Returns as list
        ```
    """
    return api_request('/workspaces/' + resolve_workspace(workspace))


@df
def create_workspace(
    display_name: str,
    *,
    capacity: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Create a new workspace with the specified name, capacity and description.

    Args:
        display_name (str): The name of the workspace to create.
        capacity (str, optional): The ID or name of the capacity to assign to the workspace. Defaults to None.
        description (str, optional): A description for the workspace. Defaults to None.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The details of the created or updated workspace if successful, otherwise None.

    Examples:
        ```python
        create_workspace(
            'MyProject',
            capacity_id='85974fbf-2c3b-4d5e-8f6a-7b8c9d0e1f2g',
            description='My new project workspace',
        )
        ```
    """
    payload = {'displayName': display_name}

    if capacity:
        payload['capacityId'] = resolve_capacity(capacity)

    if description:
        payload['description'] = description

    return api_request('/workspaces', payload=payload, method='post')


@df
def update_workspace(
    workspace: str,
    *,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of a workspace.

    Args:
        workspace (str): The workspace name or ID to update.
        display_name (str, optional): The new name for the workspace.
        description (str, optional): The new description for the workspace.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (DataFrame | dict | None): The updated workspace details. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        update_workspace(
            '123e4567-e89b-12d3-a456-426614174000',
            display_name='New Workspace Name',
        )

        update_workspace(
            'MyProjectWorkspace',
            description='Updated description',
        )

        update_workspace(
            '123e4567-e89b-12d3-a456-426614174000',
            display_name='New Workspace Name',
            description='Updated description',
        )
        ```
    """
    if not display_name and not description:
        logger.warning(
            'No changes provided. Please specify at least one property to update.'
        )
        return None

    payload = {}
    if display_name:
        payload['displayName'] = display_name
    if description:
        payload['description'] = description

    return api_request(
        '/workspaces/' + resolve_workspace(workspace),
        payload=payload,
        method='patch',
    )


def delete_workspace(workspace: str) -> None:
    """
    Delete a workspace.

    Args:
        workspace (str): The name or ID of the workspace to delete.

    Returns:
        None

    Raises:
        ResourceNotFoundError: If the specified workspace is not found.

    Examples:
        ```python
        delete_workspace('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    if not workspace_id:
        return None

    return api_request(
        '/workspaces/' + resolve_workspace(workspace),
        method='delete',
    )


@df
def list_workspace_role_assignments(
    workspace: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Lists all role assignments for a workspace.

    Args:
        workspace (str): The name or ID of the workspace to list role assignments for.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of role assignments.


    Examples:
        ```python
        list_workspace_role_assignments('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    return api_request(
        '/workspaces/' + resolve_workspace(workspace) + '/roleAssignments',
        support_pagination=True,
    )


@df
def get_workspace_role_assignment(
    workspace: str,
    user_uuid: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Retrieves the role of a user in a workspace.

    Args:
        workspace (str): The name or id of the workspace to get role assignment for.
        user_uuid (str): The UUID of the user to check.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The role assignment if found, otherwise None.

    Examples:
        ```python
        get_workspace_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'FefEFewf-feF-1234-5678-9abcdef01234'
        )
        ```
    """
    return api_request(
        '/workspaces/'
        + resolve_workspace(workspace)
        + '/roleAssignments/'
        + user_uuid
    )


@df
def add_workspace_role_assignment(
    workspace: str,
    user_uuid: str,
    user_type: Literal[
        'User', 'Group', 'ServicePrincipal', 'ServicePrincipalProfile'
    ] = 'User',
    role: Literal['Admin', 'Contributor', 'Member', 'Viewer'] = 'Admin',
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Adds a permission to a workspace for a user.

    Args:
        workspace (str): The ID or name of the workspace.
        user_uuid (str): The UUID of the user.
        user_type (str): The type of user (options: User, Group, ServicePrincipal, ServicePrincipalProfile).
        role (str): The role to assign (options: admin, member, contributor, viewer).
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The role assignment details if successful.

    Raises:
        ResourceNotFoundError: If the specified workspace is not found.
        OptionNotAvailableError: If the user type or role is invalid.

    Examples:
        ```python
        add_workspace_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'FefEFewf-feF-1234-5678-9abcdef01234', user_type='User', role='Admin'
        )
        ```
    """
    if user_type not in [
        'User',
        'Group',
        'ServicePrincipal',
        'ServicePrincipalProfile',
    ]:
        raise OptionNotAvailableError(
            f'Invalid user type: {user_type}. Must be one of: User, Group, ServicePrincipal, ServicePrincipalProfile'
        )
    if role not in ['Admin', 'Contributor', 'Member', 'Viewer']:
        raise OptionNotAvailableError(
            f'Invalid role: {role}. Must be one of: Admin, Contributor, Member, Viewer'
        )
    payload = {'principal': {'id': user_uuid, 'type': user_type}, 'role': role}

    return api_request(
        '/workspaces/' + resolve_workspace(workspace) + '/roleAssignments',
        payload=payload,
        method='post',
    )


@df
def update_workspace_role_assignment(
    workspace: str,
    user_uuid: str,
    role: Literal['Admin', 'Contributor', 'Member', 'Viewer'] = 'Admin',
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Update a role to a existing workspace role assignment.

    Args:
        workspace (str): The ID or name of the workspace.
        user_uuid (str): The UUID of the user.
        role (str): The new role to assign (options: admin, member, contributor, viewer).
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated role assignment details if successful, otherwise None.

    Raises:
        ResourceNotFoundError: If the specified workspace or user is not found.
        OptionNotAvailableError: If the role is invalid.

    Examples:
        ```python
        update_workspace_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'FefEFewf-feF-1234-5678-9abcdef01234',
            role='Contributor'
        )
    """
    if role not in ['Admin', 'Contributor', 'Member', 'Viewer']:
        raise OptionNotAvailableError(
            f'Invalid role: {role}. Must be one of: Admin, Contributor, Member, Viewer'
        )

    payload = {'role': role}

    workspace_id = resolve_workspace(workspace)

    return api_request(
        '/workspaces/' + workspace_id + '/roleAssignments/' + user_uuid,
        payload=payload,
        method='patch',
    )


def delete_workspace_role_assignment(
    workspace: str,
    user_uuid: str,
) -> None:
    """
    Removes a permission from a workspace for a user.

    Args:
        workspace (str): The ID or name of the workspace.
        user_uuid (str): The ID of the user to remove.

    Returns:
        None: If the role assignment is successfully deleted.

    Examples:
    ```python
        delete_workspace_role_assignment(
            '123e4567-e89b-12d3-a456-426614174000',
            'FefEFewf-feF-1234-5678-9abcdef01234',
        )
    ```
    """
    return api_request(
        '/workspaces/'
        + resolve_workspace(workspace)
        + '/roleAssignments/'
        + user_uuid,
        method='delete',
    )


def assign_to_capacity(workspace: str, capacity: str) -> None:
    """
    Assigns a workspace to a capacity.

    Args:
        workspace (str): The ID or name of the workspace to assign.
        capacity (str): The ID or name of the capacity to assign the workspace to.

    Returns:
        None

    Examples:
        ```python
        assign_to_capacity(
            '123e4567-e89b-12d3-a456-426614174000',
            'b7e2c1a4-8f3e-4c2a-9d2e-7b1e5f6a8c9d'
        )
        ```
    """
    payload = {'capacityId': resolve_capacity(capacity)}
    response = api_request(
        '/workspaces/' + resolve_workspace(workspace) + '/assignToCapacity',
        payload=payload,
        method='post',
        return_raw=True,
    )
    if response.status_code == 202:
        # Assuming the API returns a 202 status code for successful assignment
        logger.success(
            f'Workspace {workspace} assigned to capacity {capacity} successfully.'
        )
        return None
    else:
        logger.error(f'{response.status_code}: {response.error}')
        return None


def unassign_from_capacity(workspace: str) -> None:
    """
    Unassigns a workspace from its current capacity.

    Args:
        workspace (str): The ID or name of the workspace to unassign.

    Returns:
        None

    Examples:
        ```python
        unassign_from_capacity('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    response = api_request(
        '/workspaces/'
        + resolve_workspace(workspace)
        + '/unassignFromCapacity',
        method='post',
        return_raw=True,
    )
    if response.status_code == 202:
        logger.success(
            f'Workspace {workspace} unassigned from capacity successfully.'
        )
        return None
    else:
        logger.error(f'{response.status_code}: {response.error}')
        return None



from ._version import __version__
from .api.api import ApiResult, api_request
from .api.auth import clear_token_cache, set_auth_provider
from .cd.support_files import create_support_files
from .core.capacities import get_capacity_id, list_capacities, resolve_capacity
from .core.connections import (
    add_connection_role_assignment,
    bind_semantic_model_connection,
    create_adlsgen2_connection_with_service_principal_credentials,
    create_github_source_control_connection,
    create_sql_cloud_connection,
    create_sql_on_premises_connection,
    delete_connection,
    delete_connection_role_assignment,
    get_connection,
    get_connection_id,
    get_connection_role_assignment,
    list_connection_role_assignments,
    list_connections,
    resolve_connection,
    update_connection_role_assignment,
)
from .core.deployment_pipelines import (
    add_deployment_pipeline_role_assignment,
    assign_workspace_to_stage,
    create_deployment_pipeline,
    delete_deployment_pipeline,
    delete_deployment_pipeline_role_assignment,
    deploy_stage_content,
    get_deployment_pipeline,
    get_deployment_pipeline_id,
    get_deployment_pipeline_operation,
    list_deployment_pipeline_operations,
    list_deployment_pipeline_role_assignments,
    list_deployment_pipelines,
    resolve_deployment_pipeline,
    resolve_deployment_pipeline_stage,
    unassign_workspace_to_stage,
    update_deployment_pipeline,
    update_deployment_pipeline_stage,
)
from .core.domains import (
    assign_domain_workspaces_by_capacities,
    assign_domain_workspaces_by_ids,
    assign_domain_workspaces_by_principals,
    create_domain,
    delete_domain,
    domain_role_assignments_bulk_assign,
    domain_role_assignments_bulk_unassign,
    domain_sync_role_assignments_to_subdomain,
    get_domain,
    get_domain_id,
    list_domain_role_assignments,
    list_domain_workspaces,
    list_domains,
    resolve_domain,
    unassign_all_domain_workspaces,
    unassign_domain_workspaces_by_ids,
    update_domain,
)
from .core.folders import (
    create_folder,
    delete_folder,
    get_folder,
    get_folder_id,
    list_folders,
    move_folder,
    resolve_folder,
    update_folder,
)
from .core.gateways import (
    get_gateway,
    get_gateway_id,
    get_gateway_public_key,
    list_gateways,
    resolve_gateway,
)
from .core.git import (
    ado_connect,
    commit_to_git,
    get_git_connection,
    get_my_git_credentials,
    git_disconnect,
    git_init,
    git_status,
    github_connect,
    update_from_git,
    update_my_git_connection,
)
from .core.tags import (
    bulk_create_tags,
    delete_tag,
    get_tag_id,
    list_tags,
    resolve_tag,
    update_tag,
)
from .core.workspaces import (
    add_workspace_role_assignment,
    assign_to_capacity,
    create_workspace,
    delete_workspace,
    delete_workspace_role_assignment,
    get_workspace,
    get_workspace_id,
    get_workspace_role_assignment,
    list_workspace_role_assignments,
    list_workspaces,
    resolve_workspace,
    unassign_from_capacity,
    update_workspace,
    update_workspace_role_assignment,
)
from .dmv.dmv import (
    dmv_fetch_partitions_enriched,
    dmv_fetch_partitions_raw,
    dmv_fetch_tables_raw,
    evaluate_dmv_queries,
    import_pyadomd,
    set_dmv_connection_string_spn,
    set_dmv_connection_string_user,
)
from .graph.users import (
    get_app_registration_id,
    get_app_registration_name,
    get_security_group_id,
    get_security_group_name,
    get_service_principal_id,
    get_service_principal_name,
    get_user_email,
    get_user_id,
)
from .helpers.data_pipelines import (
    deploy_all_data_pipelines,
    deploy_data_pipeline,
    export_all_data_pipelines,
    export_data_pipeline,
    extract_data_pipeline_variables,
    get_all_data_pipelines_config,
    get_data_pipeline_config,
    replace_data_pipeline_placeholders_with_variables,
    replace_data_pipeline_variables_with_placeholders,
)
from .helpers.dataflows_gen1 import (
    deploy_all_dataflows_gen1,
    deploy_dataflow_gen1,
    export_all_dataflows_gen1,
    export_dataflow_gen1,
    get_all_dataflows_gen1_config,
    get_dataflow_gen1_config,
)
from .helpers.dataflows_gen2 import (
    deploy_all_dataflows_gen2,
    deploy_dataflow_gen2,
    export_all_dataflows_gen2,
    export_dataflow_gen2,
    extract_dataflow_gen2_variables,
    get_all_dataflows_gen2_config,
    get_dataflow_gen2_config,
    replace_dataflow_gen2_placeholders_with_parameters,
    replace_dataflow_gen2_variables_with_placeholders,
)
from .helpers.environments import (
    add_environment_external_library_from_pypi,
    deploy_all_environments,
    deploy_environment,
    export_all_environments,
    export_environment,
    get_all_environments_config,
    get_environment_config,
)
from .helpers.folders import (
    create_folders_from_path_string,
    deploy_folders,
    export_folders,
    generate_folders_paths,
    get_folders_config,
    get_folders_paths,
    resolve_folder_from_id_to_path,
)
from .helpers.items import (
    deploy_all_items,
    deploy_item,
    export_all_items,
    export_item,
)
from .helpers.lakehouses import (
    export_all_lakehouses,
    export_lakehouse,
    generate_lakehouse_shortcuts_metadata,
    get_all_lakehouses_config,
    get_lakehouse_config,
    list_valid_lakehouses,
    save_lakehouse_shortcuts_metadata,
)
from .helpers.notebooks import (
    deploy_all_notebooks,
    deploy_notebook,
    export_all_notebooks,
    export_notebook,
    extract_notebook_parameters,
    get_all_notebooks_config,
    get_notebook_config,
    replace_notebook_parameters_with_placeholders,
    replace_notebook_placeholders_with_parameters,
)
from .helpers.reports import (
    convert_report_definition_to_by_connection,
    convert_report_definition_to_by_path,
    deploy_all_reports,
    deploy_report,
    export_all_reports,
    export_report,
    extract_report_definition_pbir,
    get_all_reports_config,
    get_report_config,
)
from .helpers.semantic_models import (
    bind_semantic_model_to_gateway,
    deploy_all_semantic_models,
    deploy_semantic_model,
    execute_queries,
    export_all_semantic_models,
    export_semantic_model,
    extract_tmdl_parameters_from_semantic_model,
    get_all_semantic_models_config,
    get_semantic_model_config,
    get_semantic_model_refresh_details,
    get_semantic_model_refreshes,
    list_valid_semantic_models,
    refresh_semantic_model,
    replace_semantic_model_parameters_with_placeholders,
    replace_semantic_model_placeholders_with_parameters,
)
from .helpers.warehouses import (
    export_all_warehouses,
    export_warehouse,
    get_all_warehouses_config,
    get_warehouse_config,
    list_valid_warehouses,
)
from .helpers.workspaces import get_workspace_config
from .items.data_pipelines import (
    create_data_pipeline,
    delete_data_pipeline,
    get_data_pipeline,
    get_data_pipeline_definition,
    get_data_pipeline_id,
    list_data_pipelines,
    resolve_data_pipeline,
    update_data_pipeline,
    update_data_pipeline_definition,
)
from .items.dataflows_gen1 import (
    create_dataflow_gen1,
    delete_dataflow_gen1,
    get_dataflow_gen1,
    get_dataflow_gen1_definition,
    get_dataflow_gen1_id,
    get_dataflow_gen1_transactions,
    get_dataflows_gen1_datasources,
    list_dataflows_gen1,
    refresh_dataflow_gen1,
    resolve_dataflow_gen1,
    takeover_dataflow_gen1,
    update_dataflow_gen1,
    update_dataflow_gen1_definition,
)
from .items.dataflows_gen2 import (
    create_dataflow_gen2,
    delete_dataflow_gen2,
    get_dataflow_gen2,
    get_dataflow_gen2_definition,
    get_dataflow_gen2_id,
    list_dataflows_gen2,
    resolve_dataflow_gen2,
    update_dataflow_gen2,
    update_dataflow_gen2_definition,
)
from .items.environments import (
    create_environment,
    delete_environment,
    export_environment_external_libraries,
    get_environment,
    get_environment_definition,
    get_environment_id,
    get_environment_spark_compute,
    import_environment_external_libraries,
    list_environments,
    publish_environment,
    resolve_environment,
    update_environment,
    update_environment_definition,
    update_environment_spark_compute,
)
from .items.items import (
    create_item,
    delete_item,
    get_item,
    get_item_definition,
    get_item_id,
    list_items,
    resolve_item,
    update_item,
    update_item_definition,
)
from .items.lakehouses import (
    create_lakehouse,
    delete_lakehouse,
    get_lakehouse,
    get_lakehouse_id,
    list_lakehouses,
    resolve_lakehouse,
    update_lakehouse,
)
from .items.notebooks import (
    create_notebook,
    delete_notebook,
    get_notebook,
    get_notebook_definition,
    get_notebook_id,
    list_notebooks,
    resolve_notebook,
    update_notebook,
    update_notebook_definition,
)
from .items.reports import (
    create_report,
    delete_report,
    get_report,
    get_report_definition,
    get_report_id,
    list_reports,
    resolve_report,
    update_report,
    update_report_definition,
)
from .items.semantic_models import (
    create_semantic_model,
    delete_semantic_model,
    get_semantic_model,
    get_semantic_model_definition,
    get_semantic_model_id,
    list_semantic_models,
    resolve_semantic_model,
    update_semantic_model,
    update_semantic_model_definition,
)
from .items.shortcuts import (
    create_shortcut,
    delete_shortcut,
    get_shortcut,
    list_shortcuts,
)
from .items.spark import (
    create_workspace_custom_pool,
    delete_workspace_custom_pool,
    get_workspace_custom_pool,
    get_workspace_custom_pool_id,
    get_workspace_spark_settings,
    list_workspace_custom_pools,
    resolve_workspace_custom_pool,
    update_workspace_custom_pool,
    update_workspace_spark_settings,
)
from .items.variable_libraries import (
    create_variable_library,
    delete_variable_library,
    get_variable_library,
    get_variable_library_definition,
    get_variable_library_id,
    list_variable_libraries,
    resolve_variable_library,
    update_variable_library,
    update_variable_library_definition,
)
from .items.warehouses import (
    create_warehouse,
    delete_warehouse,
    get_warehouse,
    get_warehouse_id,
    list_warehouses,
    resolve_warehouse,
    update_warehouse,
)
from .utils.decorators import df
from .utils.exceptions import (
    AuthenticationError,
    ConfigurationError,
    FileNotFoundError,
    InvalidParameterError,
    OptionNotAvailableError,
    PyFabricOpsError,
    RequestError,
    ResourceNotFoundError,
)
from .utils.logging import (
    SUCCESS_LEVEL,
    PyFabricOpsFilter,
    PyFabricOpsFormatter,
    disable_logging,
    enable_debug_mode,
    get_logger,
    reset_logging,
    setup_logging,
)
from .utils.utils import (
    copy_to_staging,
    dataframe_to_list,
    delete_path,
    extract_display_name_from_platform,
    extract_middle_path,
    find_and_replace,
    get_current_branch,
    get_root_path,
    get_workspace_suffix,
    is_valid_uuid,
    json_to_df,
    list_paths_of_type,
    load_and_sanitize,
    pack_item_definition,
    parse_definition_report,
    parse_tmdl_parameters,
    read_json,
    unpack_item_definition,
    write_json,
    write_single_line_json,
)

__all__ = [
    'ApiResult',
    'AuthenticationError',
    'ConfigurationError',
    'FileNotFoundError',
    'InvalidParameterError',
    'OptionNotAvailableError',
    'PyFabricOpsError',
    'PyFabricOpsFilter',
    'PyFabricOpsFormatter',
    'RequestError',
    'ResourceNotFoundError',
    'SUCCESS_LEVEL',
    'add_connection_role_assignment',
    'add_deployment_pipeline_role_assignment',
    'add_environment_external_library_from_pypi',
    'add_workspace_role_assignment',
    'ado_connect',
    'api_request',
    'assign_domain_workspaces_by_capacities',
    'assign_domain_workspaces_by_ids',
    'assign_domain_workspaces_by_principals',
    'assign_to_capacity',
    'assign_workspace_to_stage',
    'bind_semantic_model_connection',
    'bind_semantic_model_to_gateway',
    'bulk_create_tags',
    'clear_token_cache',
    'commit_to_git',
    'convert_report_definition_to_by_connection',
    'convert_report_definition_to_by_path',
    'copy_to_staging',
    'create_adlsgen2_connection_with_service_principal_credentials',
    'create_data_pipeline',
    'create_dataflow_gen1',
    'create_dataflow_gen2',
    'create_deployment_pipeline',
    'create_domain',
    'create_environment',
    'create_folder',
    'create_folders_from_path_string',
    'create_github_source_control_connection',
    'create_item',
    'create_lakehouse',
    'create_notebook',
    'create_report',
    'create_semantic_model',
    'create_shortcut',
    'create_sql_cloud_connection',
    'create_sql_on_premises_connection',
    'create_support_files',
    'create_variable_library',
    'create_warehouse',
    'create_workspace',
    'create_workspace_custom_pool',
    'dataframe_to_list',
    'delete_connection',
    'delete_connection_role_assignment',
    'delete_data_pipeline',
    'delete_dataflow_gen1',
    'delete_dataflow_gen2',
    'delete_deployment_pipeline',
    'delete_deployment_pipeline_role_assignment',
    'delete_domain',
    'delete_environment',
    'delete_folder',
    'delete_item',
    'delete_lakehouse',
    'delete_notebook',
    'delete_path',
    'delete_report',
    'delete_semantic_model',
    'delete_shortcut',
    'delete_tag',
    'delete_variable_library',
    'delete_warehouse',
    'delete_workspace',
    'delete_workspace_custom_pool',
    'delete_workspace_role_assignment',
    'deploy_all_data_pipelines',
    'deploy_all_dataflows_gen1',
    'deploy_all_dataflows_gen2',
    'deploy_all_environments',
    'deploy_all_items',
    'deploy_all_notebooks',
    'deploy_all_reports',
    'deploy_all_semantic_models',
    'deploy_data_pipeline',
    'deploy_dataflow_gen1',
    'deploy_dataflow_gen2',
    'deploy_environment',
    'deploy_folders',
    'deploy_item',
    'deploy_notebook',
    'deploy_report',
    'deploy_semantic_model',
    'deploy_stage_content',
    'df',
    'disable_logging',
    'dmv_fetch_partitions_enriched',
    'dmv_fetch_partitions_raw',
    'dmv_fetch_tables_raw',
    'domain_role_assignments_bulk_assign',
    'domain_role_assignments_bulk_unassign',
    'domain_sync_role_assignments_to_subdomain',
    'enable_debug_mode',
    'evaluate_dmv_queries',
    'execute_queries',
    'export_all_data_pipelines',
    'export_all_dataflows_gen1',
    'export_all_dataflows_gen2',
    'export_all_environments',
    'export_all_items',
    'export_all_lakehouses',
    'export_all_notebooks',
    'export_all_reports',
    'export_all_semantic_models',
    'export_all_warehouses',
    'export_data_pipeline',
    'export_dataflow_gen1',
    'export_dataflow_gen2',
    'export_environment',
    'export_environment_external_libraries',
    'export_folders',
    'export_item',
    'export_lakehouse',
    'export_notebook',
    'export_report',
    'export_semantic_model',
    'export_warehouse',
    'extract_data_pipeline_variables',
    'extract_dataflow_gen2_variables',
    'extract_display_name_from_platform',
    'extract_middle_path',
    'extract_notebook_parameters',
    'extract_report_definition_pbir',
    'extract_tmdl_parameters_from_semantic_model',
    'find_and_replace',
    'generate_folders_paths',
    'generate_lakehouse_shortcuts_metadata',
    'get_all_data_pipelines_config',
    'get_all_dataflows_gen1_config',
    'get_all_dataflows_gen2_config',
    'get_all_environments_config',
    'get_all_lakehouses_config',
    'get_all_notebooks_config',
    'get_all_reports_config',
    'get_all_semantic_models_config',
    'get_all_warehouses_config',
    'get_app_registration_id',
    'get_app_registration_name',
    'get_capacity_id',
    'get_connection',
    'get_connection_id',
    'get_connection_role_assignment',
    'get_current_branch',
    'get_data_pipeline',
    'get_data_pipeline_config',
    'get_data_pipeline_definition',
    'get_data_pipeline_id',
    'get_dataflow_gen1',
    'get_dataflow_gen1_config',
    'get_dataflow_gen1_definition',
    'get_dataflow_gen1_id',
    'get_dataflow_gen1_transactions',
    'get_dataflow_gen2',
    'get_dataflow_gen2_config',
    'get_dataflow_gen2_definition',
    'get_dataflow_gen2_id',
    'get_dataflows_gen1_datasources',
    'get_deployment_pipeline',
    'get_deployment_pipeline_id',
    'get_deployment_pipeline_operation',
    'get_domain',
    'get_domain_id',
    'get_environment',
    'get_environment_config',
    'get_environment_definition',
    'get_environment_id',
    'get_environment_spark_compute',
    'get_folder',
    'get_folder_id',
    'get_folders_config',
    'get_folders_paths',
    'get_gateway',
    'get_gateway_id',
    'get_gateway_public_key',
    'get_git_connection',
    'get_item',
    'get_item_definition',
    'get_item_id',
    'get_lakehouse',
    'get_lakehouse_config',
    'get_lakehouse_id',
    'get_logger',
    'get_my_git_credentials',
    'get_notebook',
    'get_notebook_config',
    'get_notebook_definition',
    'get_notebook_id',
    'get_report',
    'get_report_config',
    'get_report_definition',
    'get_report_id',
    'get_root_path',
    'get_security_group_id',
    'get_security_group_name',
    'get_semantic_model',
    'get_semantic_model_config',
    'get_semantic_model_definition',
    'get_semantic_model_id',
    'get_semantic_model_refresh_details',
    'get_semantic_model_refreshes',
    'get_service_principal_id',
    'get_service_principal_name',
    'get_shortcut',
    'get_tag_id',
    'get_user_email',
    'get_user_id',
    'get_variable_library',
    'get_variable_library_definition',
    'get_variable_library_id',
    'get_warehouse',
    'get_warehouse_config',
    'get_warehouse_id',
    'get_workspace',
    'get_workspace_config',
    'get_workspace_custom_pool',
    'get_workspace_custom_pool_id',
    'get_workspace_id',
    'get_workspace_role_assignment',
    'get_workspace_spark_settings',
    'get_workspace_suffix',
    'git_disconnect',
    'git_init',
    'git_status',
    'github_connect',
    'import_environment_external_libraries',
    'import_pyadomd',
    'is_valid_uuid',
    'json_to_df',
    'list_capacities',
    'list_connection_role_assignments',
    'list_connections',
    'list_data_pipelines',
    'list_dataflows_gen1',
    'list_dataflows_gen2',
    'list_deployment_pipeline_operations',
    'list_deployment_pipeline_role_assignments',
    'list_deployment_pipelines',
    'list_domain_role_assignments',
    'list_domain_workspaces',
    'list_domains',
    'list_environments',
    'list_folders',
    'list_gateways',
    'list_items',
    'list_lakehouses',
    'list_notebooks',
    'list_paths_of_type',
    'list_reports',
    'list_semantic_models',
    'list_shortcuts',
    'list_tags',
    'list_valid_lakehouses',
    'list_valid_semantic_models',
    'list_valid_warehouses',
    'list_variable_libraries',
    'list_warehouses',
    'list_workspace_custom_pools',
    'list_workspace_role_assignments',
    'list_workspaces',
    'load_and_sanitize',
    'move_folder',
    'pack_item_definition',
    'parse_definition_report',
    'parse_tmdl_parameters',
    'publish_environment',
    'read_json',
    'refresh_dataflow_gen1',
    'refresh_semantic_model',
    'replace_data_pipeline_placeholders_with_variables',
    'replace_data_pipeline_variables_with_placeholders',
    'replace_dataflow_gen2_placeholders_with_parameters',
    'replace_dataflow_gen2_variables_with_placeholders',
    'replace_notebook_parameters_with_placeholders',
    'replace_notebook_placeholders_with_parameters',
    'replace_semantic_model_parameters_with_placeholders',
    'replace_semantic_model_placeholders_with_parameters',
    'reset_logging',
    'resolve_capacity',
    'resolve_connection',
    'resolve_data_pipeline',
    'resolve_dataflow_gen1',
    'resolve_dataflow_gen2',
    'resolve_deployment_pipeline',
    'resolve_deployment_pipeline_stage',
    'resolve_domain',
    'resolve_environment',
    'resolve_folder',
    'resolve_folder_from_id_to_path',
    'resolve_gateway',
    'resolve_item',
    'resolve_lakehouse',
    'resolve_notebook',
    'resolve_report',
    'resolve_semantic_model',
    'resolve_tag',
    'resolve_variable_library',
    'resolve_warehouse',
    'resolve_workspace',
    'resolve_workspace_custom_pool',
    'save_lakehouse_shortcuts_metadata',
    'set_auth_provider',
    'set_dmv_connection_string_spn',
    'set_dmv_connection_string_user',
    'setup_logging',
    'takeover_dataflow_gen1',
    'unassign_all_domain_workspaces',
    'unassign_domain_workspaces_by_ids',
    'unassign_from_capacity',
    'unassign_workspace_to_stage',
    'unpack_item_definition',
    'update_connection_role_assignment',
    'update_data_pipeline',
    'update_data_pipeline_definition',
    'update_dataflow_gen1',
    'update_dataflow_gen1_definition',
    'update_dataflow_gen2',
    'update_dataflow_gen2_definition',
    'update_deployment_pipeline',
    'update_deployment_pipeline_stage',
    'update_domain',
    'update_environment',
    'update_environment_definition',
    'update_environment_spark_compute',
    'update_folder',
    'update_from_git',
    'update_item',
    'update_item_definition',
    'update_lakehouse',
    'update_my_git_connection',
    'update_notebook',
    'update_notebook_definition',
    'update_report',
    'update_report_definition',
    'update_semantic_model',
    'update_semantic_model_definition',
    'update_tag',
    'update_variable_library',
    'update_variable_library_definition',
    'update_warehouse',
    'update_workspace',
    'update_workspace_custom_pool',
    'update_workspace_role_assignment',
    'update_workspace_spark_settings',
    'write_json',
    'write_single_line_json',
]



from typing import Any, Dict, List, Literal, Optional, Union

from pandas import DataFrame
from requests import get

from ..api.api import api_request
from ..core.folders import resolve_folder
from ..core.workspaces import resolve_workspace
from ..items.spark import get_workspace_custom_pool
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_environments(
    workspace: str, *, df: Optional[bool] = True
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns a list of environments from the specified workspace.
    This API supports pagination.

    Args:
        workspace (str): The workspace name or ID.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of environments, excluding those that start with the specified prefixes. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        list_environments('MyProjectWorkspace')
        list_environments('MyProjectWorkspace', df=False)
        ```
    """
    workspace_id = resolve_workspace(workspace)
    return api_request(
        endpoint='/workspaces/' + workspace_id + '/environments',
        support_pagination=True,
    )


def get_environment_id(workspace: str, environment: str) -> str | None:
    """
    Retrieves the ID of a specific environment in the workspace.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name of the environment.

    Returns:
        str|None: The ID of the environment, or None if not found.

    Examples:
        ```python
        get_environment_id('MyProjectWorkspace', 'MyEnvironment')
        get_environment_id('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    environments = list_environments(
        workspace=resolve_workspace(workspace),
        df=False,
    )

    for _environment in environments:
        if _environment['displayName'] == environment:
            return _environment['id']
    logger.warning(
        f"environment '{environment}' not found in workspace '{workspace}'."
    )
    return None


def resolve_environment(
    workspace: str,
    environment: str,
) -> Union[str, None]:
    """
    Resolves a environment name to its ID.

    Args:
        workspace (str): The ID of the workspace.
        environment (str): The name of the environment.

    Returns:
        (Union[str, None]): The ID of the environment, or None if not found.

    Examples:
        ```python
        resolve_environment('MyProjectWorkspace', 'MyEnvironment')
        resolve_environment('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    if is_valid_uuid(environment):
        return environment
    else:
        return get_environment_id(resolve_workspace(workspace), environment)


@df
def get_environment(
    workspace: str,
    environment: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Retrieves a specific environment from the workspace.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment to retrieve.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The environment details as a dictionary or DataFrame, or None if not found.

    Examples:
        ```python
        get_environment('MyProjectWorkspace', 'MyEnvironment')
        get_environment('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', df=True)
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id,
    )


@df
def update_environment(
    workspace: str,
    environment: str,
    *,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of the specified environment.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment to update.
        display_name (str, optional): The new display name for the environment.
        description (str, optional): The new description for the environment.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated environment details if successful, otherwise None.

    Examples:
        ```python
        update_environment('MyProjectWorkspace', 'MyEnvironment', display_name='UpdatedMyEnvironment')
        update_environment('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    payload = {}

    if display_name:
        payload['displayName'] = display_name

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id,
        method='patch',
        payload=payload,
    )


def delete_environment(workspace: str, environment: str) -> None:
    """
    Delete a environment from the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace to delete.
        environment (str): The name or ID of the environment to delete.

    Returns:
        None: If the environment is successfully deleted.

    Raises:
        ResourceNotFoundError: If the specified workspace is not found.

    Examples:
        ```python
        delete_environment('MyProjectWorkspace', 'MyEnvironment')
        delete_environment('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id,
        method='delete',
    )


def get_environment_definition(
    workspace: str, environment: str
) -> Union[Dict[str, Any], None]:
    """
    Retrieves the definition of a environment by its name or ID from the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment.

    Returns:
        (Union[Dict[str, Any], None]): The environment definition if found, otherwise None.

    Examples:
        ```python
        get_environment_definition('MyProjectWorkspace', 'MyEnvironment')
        get_environment_definition('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/getDefinition',
        method='post',
        support_lro=True,
    )


@df
def update_environment_definition(
    workspace: str,
    environment: str,
    environment_definition: Dict[str, Any],
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the definition of an existing environment in the specified workspace.
    If the environment does not exist, it returns None.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment to update.
        environment_definition (Dict[str, Any]): The updated environment definition.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated environment details if successful, otherwise None.

    Examples:
        ```python
        update_environment_definition(
            'MyProjectWorkspace',
            'MyEnvironment',
            environment_definition = {...}  # Updated environment definition
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    payload = {'definition': environment_definition}

    parms = {'updateMetadata': True}

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/updateDefinition',
        payload=payload,
        params=parms,
        method='post',
        support_lro=True,
    )


@df
def create_environment(
    workspace: str,
    display_name: str,
    *,
    environment_definition: Dict[str, Any] = None,
    description: Optional[str] = None,
    folder: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Creates a new environment in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        display_name (str): The display name of the environment.
        environment_definition (Dict[str, Any]): The environment definition. If Not provided, an empty definition will be used.
        description (str, optional): A description for the environment.
        folder (str, optional): The folder to create the environment in.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created environment details.

    Examples:
        ```python
        create_environment(
            'MyProjectWorkspace', 'MyEnvironment', environment_definition={...}
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {
        'displayName': display_name,
        'definition': environment_definition,
    }

    if description:
        payload['description'] = description

    if folder:
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id:
            payload['folderId'] = folder_id

    if environment_definition:
        payload['definition'] = environment_definition

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/environments',
        method='post',
        payload=payload,
        support_lro=True,
    )


@df
def publish_environment(
    workspace: str,
    environment: str,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Trigger an environment publish operation.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment to publish.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The published environment details if successful, otherwise None.

    Examples:
        ```python
        publish_environment(
            'MyProjectWorkspace',
            'MyEnvironment'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    params = {'beta': False}

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/staging/publish',
        method='post',
        params=params,
    )


@df
def get_environment_spark_compute(
    workspace: str, environment: str
) -> Union[Dict[str, Any], None]:
    """
    Get environment staging spark compute.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment.

    Returns:
        (Union[Dict[str, Any], None]): The environment definition if found, otherwise None.

    Examples:
        ```python
        get_environment_spark_compute('MyProjectWorkspace', 'MyEnvironment')
        get_environment_spark_compute('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/staging/sparkcompute',
        params={'beta': False},
    )


@df
def update_environment_spark_compute(
    workspace: str,
    environment: str,
    *,
    pool: str = None,
    driver_cores: Literal[4, 8] = None,
    driver_memory: Literal['28g', '56g'] = None,
    executor_cores: Literal[4, 8] = None,
    executor_memory: Literal['28g', '56g'] = None,
    dynamic_executor_allocation_enabled: bool = None,
    min_executors: int = None,
    max_executors: int = None,
    spark_properties: List[Dict[str, str]] = None,
    runtime_version: Literal['1.2', '1.3', '2.0'] = None,
) -> Union[Dict[str, Any], None]:
    """
    Update environment staging spark compute.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment.

    Returns:
        (Union[Dict[str, Any], None]): The environment definition if found, otherwise None.

    Examples:
        ```python
        update_environment_spark_compute(
            'MyProjectWorkspace',
            'MyEnvironment',
            pool='Custom Pool Name',
        )
        update_environment_spark_compute(
            'MyProjectWorkspace',
            'MyEnvironment',
            pool='Custom Pool Name',
            driver_cores=8,
            driver_memory='56g',
            executor_cores=8,
            executor_memory='56g',
            dynamic_executor_allocation_enabled=True,
            min_executors=1,
            max_executors=9,
            spark_properties=[
                {"key": "spark.sql.caseSensitive", "value": "true"}
            ],
            runtime_version='1.3'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    if pool:
        _pool = get_workspace_custom_pool(workspace_id, pool, df=False)
        if not _pool:
            return None
        pool_name = _pool['name']
        pool_id = _pool['id']
    else:
        pool_name = 'Starter Pool'
        pool_id = '00000000-0000-0000-0000-000000000000'

    payload = {
        'instancePool': {'name': pool_name, 'type': 'Workspace', 'id': pool_id}
    }

    if driver_cores is not None:
        payload['driverCores'] = driver_cores

    if driver_memory is not None:
        payload['driverMemory'] = driver_memory

    if executor_cores is not None:
        payload['executorCores'] = executor_cores

    if executor_memory is not None:
        payload['executorMemory'] = executor_memory

    if dynamic_executor_allocation_enabled is not None:
        payload['dynamicExecutorAllocation'] = {
            'enabled': dynamic_executor_allocation_enabled,
            'minExecutors': min_executors,
            'maxExecutors': max_executors,
        }

    if spark_properties is not None:
        payload['sparkProperties'] = spark_properties
    if runtime_version is not None:
        payload['runtimeVersion'] = runtime_version

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/staging/sparkcompute',
        params={'beta': False},
        method='patch',
        payload=payload,
    )


def export_environment_external_libraries(
    workspace: str,
    environment: str,
) -> Union[Dict[str, Any], None]:
    """
    Export environment external libraries.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment.

    Returns:
        (Union[Dict[str, Any], None]): The environment external libraries if found, otherwise None.

    Examples:
        ```python
        export_environment_external_libraries('MyProjectWorkspace', 'MyEnvironment')
        export_environment_external_libraries('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/libraries/exportExternalLibraries',
        return_raw=True,
    ).text


def import_environment_external_libraries(
    workspace: str,
    environment: str,
) -> Union[Dict[str, Any], None]:
    """
    Import external libraries from environment.yaml in the environment.

    Args:
        workspace (str): The workspace name or ID.
        environment (str): The name or ID of the environment.

    Returns:
        (Union[Dict[str, Any], None]): The environment external libraries if found, otherwise None.

    Examples:
        ```python
        import_environment_external_libraries('MyProjectWorkspace', 'MyEnvironment')
        import_environment_external_libraries('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    environment_id = resolve_environment(workspace_id, environment)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/environments/'
        + environment_id
        + '/staging/libraries/importExternalLibraries',
        method='post',
    )



import json
import os
import tempfile
import time
from abc import ABC, abstractmethod
from typing import Dict, Literal, Optional, Union

import requests
from azure.identity import InteractiveBrowserCredential
from dotenv import load_dotenv

from ..utils.exceptions import (
    AuthenticationError,
    OptionNotAvailableError,
    ResourceNotFoundError,
)
from ..utils.logging import get_logger
from .scopes import FABRIC_SCOPE, GRAPH_SCOPE, POWERBI_SCOPE, TOKEN_TEMPLATE

logger = get_logger(__name__)

# Define what should be publicly exported from this module
__all__ = ['set_auth_provider', 'clear_token_cache']


class TokenCache:
    """Manage the token cache in a temporary file"""

    CACHE_TEMPLATE = {
        'FABRIC_SPN': {'access_token': '', 'expires_at': 0},
        'FABRIC_USER': {'access_token': '', 'expires_at': 0},
        'FABRIC_INTERACTIVE': {'access_token': '', 'expires_at': 0},
        'FABRIC_NOTEBOOK': {'access_token': '', 'expires_at': 0},
        'POWERBI_SPN': {'access_token': '', 'expires_at': 0},
        'POWERBI_USER': {'access_token': '', 'expires_at': 0},
        'POWERBI_INTERACTIVE': {'access_token': '', 'expires_at': 0},
        'POWERBI_NOTEBOOK': {'access_token': '', 'expires_at': 0},
        'GRAPH_SPN': {'access_token': '', 'expires_at': 0},
        'GRAPH_USER': {'access_token': '', 'expires_at': 0},
        'GRAPH_INTERACTIVE': {'access_token': '', 'expires_at': 0},
        'GRAPH_NOTEBOOK': {'access_token': '', 'expires_at': 0},
    }

    def __init__(self, cache_file: Optional[str] = None):
        self.cache_file = cache_file or os.path.join(
            tempfile.gettempdir(), 'pf_token_cache.json'
        )
        self._init_cache()

    def _init_cache(self):
        """Initialize the cache file if it does not exist"""
        if not os.path.exists(self.cache_file):
            with open(self.cache_file, 'w') as f:
                json.dump(self.CACHE_TEMPLATE, f)

    def load_tokens(self) -> Dict:
        """Load tokens from cache"""
        try:
            with open(self.cache_file, 'r') as f:
                return json.load(f)
        except (FileNotFoundError, json.JSONDecodeError):
            self._init_cache()
            return self.CACHE_TEMPLATE.copy()

    def save_tokens(self, tokens: Dict):
        """Save tokens to cache"""
        with open(self.cache_file, 'w') as f:
            json.dump(tokens, f, indent=4)

    def get_token(self, token_key: str) -> Optional[Dict]:
        """Get a specific token from cache"""
        tokens = self.load_tokens()
        return tokens.get(token_key)

    def is_token_valid(
        self, token_key: str, buffer_seconds: int = 300
    ) -> bool:
        """Check if a token is still valid"""
        token_data = self.get_token(token_key)
        if not token_data or not token_data.get('access_token'):
            return False

        now = time.time()
        expires_at = token_data.get('expires_at', 0)
        return (expires_at - now) > buffer_seconds

    def store_token(self, token_key: str, access_token: str, expires_in: int):
        """Store a new token in cache"""
        tokens = self.load_tokens()
        tokens[token_key] = {
            'access_token': access_token,
            'expires_at': time.time() + expires_in,
        }
        self.save_tokens(tokens)

    def clear_cache(self):
        """Clear the token cache by deleting the cache file"""
        if os.path.exists(self.cache_file):
            os.remove(self.cache_file)
            logger.info(f'Token cache cleared: {self.cache_file}')
        else:
            logger.warning(f'Cache file not found: {self.cache_file}')


class CredentialProvider(ABC):
    """Abstract class for different credential providers"""

    @abstractmethod
    def get_credentials(self) -> Dict[str, str]:
        """Return the necessary credentials"""
        pass


class EnvCredentialProvider(CredentialProvider):
    """Environment variable credential provider"""

    def get_credentials(self) -> Dict[str, str]:
        load_dotenv()
        return {
            'fab_client_id': os.getenv('FAB_CLIENT_ID'),
            'fab_client_secret': os.getenv('FAB_CLIENT_SECRET'),
            'fab_tenant_id': os.getenv('FAB_TENANT_ID'),
            'fab_username': os.getenv('FAB_USERNAME'),
            'fab_password': os.getenv('FAB_PASSWORD'),
            'github_token': os.getenv('GH_TOKEN'),
        }


class OAuthProvider:
    """OAuth interactive authentication provider"""

    def __init__(self, cache: TokenCache):
        self.cache = cache

    def get_token(
        self, audience: Literal['fabric', 'powerbi', 'graph'] = 'fabric'
    ) -> Dict:
        if audience not in ['fabric', 'powerbi', 'graph']:
            raise OptionNotAvailableError(
                f'Audience not available. Available: fabric, powerbi, graph. Got: {audience}'
            )
        if audience == 'graph':
            scope == GRAPH_SCOPE
        elif audience == 'powerbi':
            scope = POWERBI_SCOPE
        else:
            scope = FABRIC_SCOPE
        token_key = f'{audience.upper()}_INTERACTIVE'

        # Check if cached token is still valid
        if self.cache.is_token_valid(token_key):
            return self.cache.get_token(token_key)

        logger.info('Opening browser for user authentication...')
        credential = InteractiveBrowserCredential()
        new_token = credential.get_token(scope)

        if not new_token:
            raise ResourceNotFoundError('Access token not found.')

        logger.success('Token retrieved successfully.')

        # Calculate expires_in based on expires_on
        expires_in = int(new_token.expires_on - time.time())
        self.cache.store_token(token_key, new_token.token, expires_in)

        return self.cache.get_token(token_key)


class FabricNotebookProvider:
    """Fabric Notebook authentication provider using notebookutils.credentials"""

    def __init__(self, cache: TokenCache):
        self.cache = cache
        self._notebookutils = None

    def _get_notebookutils(self):
        """Lazy load notebookutils module"""
        if self._notebookutils is None:
            try:
                from notebookutils import credentials

                self._notebookutils = credentials
            except ImportError:
                raise AuthenticationError(
                    'notebookutils is not available. '
                    'This authentication method only works inside Microsoft Fabric notebooks. '
                    'If you are running outside Fabric, use set_auth_provider("env") or set_auth_provider("oauth") instead.'
                )
        return self._notebookutils

    def get_token(
        self, audience: Literal['fabric', 'powerbi', 'graph'] = 'fabric'
    ) -> Dict:
        """Get token from Fabric notebook context"""
        token_key = f'{audience.upper()}_NOTEBOOK'

        # Check if cached token is still valid
        if self.cache.is_token_valid(token_key):
            return self.cache.get_token(token_key)

        logger.info('Getting token from Fabric notebook context...')
        credentials = self._get_notebookutils()

        # Get token using notebookutils
        # For Power BI API, use 'pbi' resource
        # For Fabric API, use 'storage' or the appropriate resource
        resource = 'pbi' if audience == 'powerbi' else 'pbi'
        access_token = credentials.getToken(resource)

        if not access_token:
            raise ResourceNotFoundError(
                f'Access token not found for resource: {resource}'
            )

        logger.success('Token retrieved successfully from Fabric notebook.')

        # Store in cache with default expiration (1 hour)
        expires_in = 3600
        self.cache.store_token(token_key, access_token, expires_in)

        return self.cache.get_token(token_key)


class TokenManager:
    """Main token and authentication manager"""

    def __init__(
        self, auth_provider: Literal['env', 'oauth', 'fabric'] = 'env'
    ):
        self.cache = TokenCache()
        self.auth_provider = auth_provider
        self._credential_providers = {
            'env': EnvCredentialProvider(),
        }
        self.oauth_provider = OAuthProvider(self.cache)
        self.fabric_provider = FabricNotebookProvider(self.cache)

    def set_auth_provider(
        self, source: Literal['env', 'oauth', 'fabric'] = 'env'
    ):
        """Define the authentication provider"""
        if source not in ['env', 'oauth', 'fabric']:
            raise OptionNotAvailableError(
                f'Source not available. Available: env, oauth, fabric. Got: {source}'
            )
        self.auth_provider = source

    def _build_token_payload(
        self,
        audience: Literal['fabric', 'powerbi', 'graph'],
        credential_type: Literal['spn', 'user'],
        credentials: Dict[str, str],
    ) -> Dict:
        """Construct the payload for token request"""
        if audience == 'graph':
            scope = GRAPH_SCOPE
        elif audience == 'powerbi':
            scope = POWERBI_SCOPE
        else:
            scope = FABRIC_SCOPE

        payload = {
            'client_id': credentials['fab_client_id'],
            'client_secret': credentials['fab_client_secret'],
            'tenant_id': credentials['fab_tenant_id'],
            'grant_type': 'client_credentials'
            if credential_type == 'spn'
            else 'password',
            'scope': scope,
        }

        if credential_type == 'user':
            payload['username'] = credentials['fab_username']
            payload['password'] = credentials['fab_password']

        return payload

    def _retrieve_token_from_api(
        self,
        audience: Literal['fabric', 'powerbi', 'graph'],
        credential_type: Literal['spn', 'user'],
    ) -> Dict:
        """Makes an HTTP request to retrieve the token"""
        if self.auth_provider not in self._credential_providers:
            raise OptionNotAvailableError(
                f'Invalid auth provider: {self.auth_provider}'
            )

        credentials = self._credential_providers[
            self.auth_provider
        ].get_credentials()
        tenant_id = credentials['fab_tenant_id']
        url = TOKEN_TEMPLATE.format(tenant_id=tenant_id)

        payload = self._build_token_payload(
            audience, credential_type, credentials
        )

        try:
            resp = requests.post(url, data=payload)
            if resp.status_code == 200:
                return resp.json()
            else:
                raise AuthenticationError(
                    f'Token request failed: {resp.status_code} - {resp.text}'
                )
        except Exception as e:
            raise AuthenticationError(f'Failed to retrieve token: {str(e)}')

    def get_token(
        self,
        audience: Literal['fabric', 'powerbi', 'graph'] = 'fabric',
        credential_type: Literal['spn', 'user'] = 'spn',
    ) -> Dict:
        """Get a valid token, using cache when possible"""

        # OAuth uses a different flow
        if self.auth_provider == 'oauth':
            return self.oauth_provider.get_token(audience)

        # Fabric notebook uses notebookutils
        if self.auth_provider == 'fabric':
            return self.fabric_provider.get_token(audience)

        # For env, use cache + API
        token_key = f'{audience.upper()}_{credential_type.upper()}'

        # Check if cached token is still valid
        if self.cache.is_token_valid(token_key):
            return self.cache.get_token(token_key)

        # Fetch new token from API
        token_response = self._retrieve_token_from_api(
            audience, credential_type
        )
        if not token_response:
            raise ResourceNotFoundError('Access token not found.')

        # Store in cache
        self.cache.store_token(
            token_key,
            token_response['access_token'],
            token_response['expires_in'],
        )

        return self.cache.get_token(token_key)


# Global instance of the token manager
# This allows the same instance to be used across the application
_token_manager = TokenManager()


def set_auth_provider(
    source: Literal['env', 'oauth', 'fabric'] = 'env'
) -> None:
    """
    Set the authentication provider for token retrieval.

    Args:
        source (str): The provider of credentials. Can be "env", "oauth", or "fabric".

    Returns:
        None

    Raises:
        OptionNotAvailableError: If the source is not one of the available options.

    Examples:
        ### Environment variables (.env, GitHub Secrets, Ado Secrets...)
        ```python
        set_auth_provider("env")
        ```

        ### OAuth (Interactive)
        ```python
        set_auth_provider("oauth")
        ```

        ### Fabric Notebook (uses notebookutils.credentials)
        ```python
        set_auth_provider("fabric")
        ```
        This method is designed for use inside Microsoft Fabric notebooks where
        the user is already authenticated. It uses notebookutils.credentials.getToken()
        to retrieve the access token.
    """
    global _token_manager
    _token_manager.set_auth_provider(source)


def clear_token_cache() -> None:
    """
    Clear the token cache by deleting the cache file.

    This will force all subsequent token requests to retrieve new tokens
    from the authentication provider.

    Returns:
        None

    Examples:
        ```python
        from pyfabricops.api.auth import clear_token_cache

        # Clear all cached tokens
        clear_token_cache()
        ```
    """
    global _token_manager
    _token_manager.cache.clear_cache()


def _get_token(
    audience: Literal['fabric', 'powerbi', 'graph'] = 'fabric',
    auth_provider: Literal['env', 'oauth', 'fabric'] = 'env',
    credential_type: Literal['spn', 'user'] = 'spn',
) -> Union[dict, None]:
    """Get a token"""
    return _token_manager.get_token(audience, credential_type)



import time
from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.folders import resolve_folder
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_lakehouses(
    workspace: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns a list of lakehouses from the specified workspace.
    This API supports pagination.

    Args:
        workspace (str): The workspace name or ID.
        ddf (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of lakehouses, excluding those that start with the specified prefixes. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        list_lakehouses('MyProjectWorkspace')
        ```
    """
    return api_request(
        endpoint='/workspaces/' + resolve_workspace(workspace) + '/lakehouses',
        support_pagination=True,
    )


def get_lakehouse_id(workspace: str, lakehouse: str) -> Union[str, None]:
    """
    Retrieves the ID of a lakehouse by its name from the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        lakehouse (str): The name of the lakehouse.

    Returns:
        (Union[str, None]): The ID of the lakehouse, or None if not found.

    Examples:
        ```python
        get_lakehouse_id('MyProjectWorkspace', 'SalesDataLakehouse')
        ```
    """
    lakehouses = list_lakehouses(workspace, df=False)
    if not lakehouses:
        return None

    for lakehouse_ in lakehouses:
        if lakehouse_['displayName'] == lakehouse:
            return lakehouse_['id']

    logger.warning(
        f'Lakehouse {lakehouse} not found in workspace {workspace}.'
    )
    return None


def resolve_lakehouse(
    workspace: str,
    lakehouse: str,
) -> Union[str, None]:
    """
    Resolves a lakehouse name to its ID.

    Args:
        workspace (str): The ID of the workspace.
        lakehouse (str): The name of the lakehouse.

    Returns:
        (Union[str, None]): The ID of the lakehouse, or None if not found.

    Examples:
        ```python
        resolve_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse')
        ```
    """
    if is_valid_uuid(lakehouse):
        return lakehouse
    else:
        return get_lakehouse_id(workspace, lakehouse)


@df
def get_lakehouse(
    workspace: str,
    lakehouse: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Retrieves a lakehouse by its name or ID from the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        lakehouse (str): The name or ID of the lakehouse.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The lakehouse details if found. If `df=True`, returns a DataFrame with flattened keys.

    Examples:
        ```python
        get_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse')
        get_lakehouse('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        get_lakehouse('123e4567-e89b-12d3-a456-426614174000', 'SalesDataLakehouse', df=True)
        ```
    """
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    lakehouse_id = resolve_lakehouse(workspace_id, lakehouse)
    if not lakehouse_id:
        return None

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/lakehouses/' + lakehouse_id,
    )


@df
def create_lakehouse(
    workspace: str,
    display_name: str,
    *,
    description: Optional[str] = None,
    folder: Optional[str] = None,
    enable_schemas: Optional[bool] = False,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Create a lakehouse in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        display_name (str): The display name for the lakehouse.
        description (Optional[str]): The description for the lakehouse.
        folder (Optional[str]): The folder to create the lakehouse in.
        enable_schemas (Optional[bool]): Whether to enable schemas for the lakehouse.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created lakehouse details if successful, otherwise None.

    Examples:
        ```python
        create_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse')
        create_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse', description='Sales data lakehouse')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {'displayName': display_name}

    if description:
        payload['description'] = description

    if folder:
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id:
            payload['folderId'] = folder_id

    if enable_schemas:
        payload['creationPayload'] = {'enableSchemas': True}

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/lakehouses',
        method='post',
        payload=payload,
    )


@df
def update_lakehouse(
    workspace: str,
    lakehouse: str,
    *,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of the specified lakehouse.

    Args:
        workspace (str): The workspace name or ID.
        lakehouse (str): The name or ID of the lakehouse to update.
        display_name (Optional[str]): The new display name for the lakehouse.
        description (Optional[str]): The new description for the lakehouse.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated lakehouse details if successful, otherwise None.

    Examples:
        ```python
        update_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse', display_name='UpdatedSalesDataLakehouse')
        update_lakehouse('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    lakehouse_id = resolve_lakehouse(workspace_id, lakehouse)

    payload = {}

    if display_name:
        payload['displayName'] = display_name

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/lakehouses/' + lakehouse_id,
        method='patch',
        payload=payload,
    )


def delete_lakehouse(workspace: str, lakehouse: str) -> None:
    """
    Delete a lakehouse in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        lakehouse (str): The name or ID of the lakehouse to delete.

    Returns:
        (bool): True if the lakehouse was deleted successfully, otherwise False.

    Examples:
        ```python
        delete_lakehouse('MyProjectWorkspace', 'SalesDataLakehouse')
        delete_lakehouse('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    lakehouse_id = resolve_lakehouse(workspace_id, lakehouse)
    return api_request(
        endpoint='/workspaces/' + workspace_id + '/lakehouses/' + lakehouse_id,
        method='delete',
    )



import glob
import json
import os
import re
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd
from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import (
    create_folders_from_path_string,
    resolve_folder_from_id_to_path,
)
from ..items.reports import (
    create_report,
    get_report,
    get_report_definition,
    list_reports,
    resolve_report,
    update_report_definition,
)
from ..items.semantic_models import get_semantic_model_id
from ..utils.decorators import df
from ..utils.exceptions import ResourceNotFoundError
from ..utils.logging import get_logger
from ..utils.utils import (
    extract_display_name_from_platform,
    extract_middle_path,
    list_paths_of_type,
    pack_item_definition,
    unpack_item_definition,
)

logger = get_logger(__name__)


def get_report_config(
    workspace: str, report: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific report config from a workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        report (str): The name or ID of the report.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the report.
    """
    item = report
    item_data = get_report(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('displayName')] = {}

        config = {
            'id': item_data['id'],
            'description': item_data.get('description', None),
            'folder_id': ''
            if item_data.get('folderId') is None
            or pd.isna(item_data.get('folderId'))
            else item_data['folderId'],
        }

        return config


def get_all_reports_config(
    workspace: str,
) -> Union[Dict[str, Any], None]:
    """
    Get reports config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config of all reports in the workspace
    """
    items = list_reports(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:

        item_data = get_report(workspace, item['id'], df=False)

        config[item['displayName']] = {
            'id': item['id'],
            'description': item.get('description', None),
            'folder_id': ''
            if item.get('folderId') is None or pd.isna(item.get('folderId'))
            else item['folderId'],
        }

    return config


def export_report(
    workspace: str,
    report: str,
    path: Union[str, Path],
) -> None:
    """
    Export a report to path.

    Args:
        workspace (str): The name or ID of the workspace.
        report (str): The name or ID of the report.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    item = get_report(workspace_id, report, df=False)
    try:
        folder_path = resolve_folder_from_id_to_path(
            workspace_id, item['folderId']
        )
    except:
        logger.info(f'{item["displayName"]}.Report is not inside a folder.')
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + '.Report')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + '.Report')
        )
    os.makedirs(item_path, exist_ok=True)

    definition = get_report_definition(workspace_id, item['id'])
    if definition is None:
        return None

    unpack_item_definition(definition, item_path)

    logger.success(
        f'`{item["displayName"]}.Report` was exported to {item_path} successfully.'
    )
    return None


def export_all_reports(
    workspace: str,
    path: Union[str, Path],
) -> None:
    """
    Export a report to path.

    Args:
        workspace (str): The name or ID of the workspace.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_reports(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, item['folderId']
            )
        except:
            logger.info(
                f'{item["displayName"]}.Report is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + '.Report')
        else:
            item_path = (
                Path(path) / folder_path / (item['displayName'] + '.Report')
            )
        os.makedirs(item_path, exist_ok=True)

        definition = get_report_definition(workspace_id, item['id'])
        if definition is None:
            return None

        unpack_item_definition(definition, item_path)

    logger.success(f'All reports were exported to {path} successfully.')
    return None


def extract_report_definition_pbir(path: Union[str, Path]) -> Dict[str, str]:
    """
    Parse a Power BI report definition file to extract workspace and semantic model information.

    Args:
        path (str): The path to the Power BI report definition file.

    Returns:
        dict: A dictionary containing the workspace name, semantic model name, and semantic model ID.

    Raises:
        ResourceNotFoundError: If the specified file does not exist or is not in the expected format.
        FileNotFoundError: If the specified file does not exist.
        json.JSONDecodeError: If the file content is not valid JSON.

    Examples:
        ```python
        parse_definition_report('MyProject/workspace/path/to/Financials.Report/definition.pbir')
        ```
    """
    path = Path(path) / 'definition.pbir'
    with open(path, encoding='utf-8') as f:
        data = json.load(f)

    # Check if the file contains the expected structure
    if (
        'datasetReference' not in data
        or 'byConnection' not in data['datasetReference']
    ):
        raise ResourceNotFoundError(
            f'Invalid report definition file: {path}. Expected structure not found.'
        )
    by_conn = data['datasetReference']['byConnection']
    conn_str = by_conn['connectionString']
    # model_id = by_conn['pbiModelDatabaseName']

    # 1) workspace_name: part after the last slash and before the semicolon
    workspace_name = conn_str.split('/')[-1].split(';')[0]

    # 2) semantic_model_name: "initial catalog" value from the connection string
    m = re.search(r'initial catalog=([^;]+)', conn_str, re.IGNORECASE)
    semantic_model_name = m.group(1) if m else None

    # 3) semantic_model_id: takes directly from the pbiModelDatabaseName field
    m_id = re.search(r'semanticmodelid=([^;]+)', conn_str, re.IGNORECASE)
    semantic_model_id = m_id.group(1) if m_id else None

    return {
        'workspace_name': workspace_name,
        'semantic_model_name': semantic_model_name,
        'semantic_model_id': semantic_model_id,
    }


@df
def deploy_report(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Deploy a report to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the report.
        start_path (Optional[str]): The starting path for folder creation.
        description (Optional[str]): Description for the report.
        df (Optional[bool]): If True, returns a DataFrame, otherwise returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The deployed report or None if deployment fails.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    display_name = extract_display_name_from_platform(path)
    if display_name is None:
        return None

    item_id = resolve_report(workspace_id, display_name)

    folder_path_string = extract_middle_path(path, start_path=start_path)
    folder_id = create_folders_from_path_string(
        workspace_id, folder_path_string
    )

    item_definition = pack_item_definition(path)

    if item_id is None:
        return create_report(
            workspace_id,
            display_name=display_name,
            item_definition=item_definition,
            description=description,
            folder=folder_id,
            df=False,
        )

    else:
        return update_report_definition(
            workspace_id,
            item_id,
            item_definition=item_definition,
            df=False,
        )


def deploy_all_reports(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all reports to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the reports.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    reports_paths = list_paths_of_type(path, 'Report')

    for path_ in reports_paths:

        display_name = extract_display_name_from_platform(path_)
        if display_name is None:
            return None

        item_id = resolve_report(workspace_id, display_name)

        folder_path_string = extract_middle_path(path_, start_path=start_path)
        folder_id = create_folders_from_path_string(
            workspace_id, folder_path_string
        )

        item_definition = pack_item_definition(path_)

        if item_id is None:
            create_report(
                workspace_id,
                display_name=display_name,
                item_definition=item_definition,
                folder=folder_id,
                df=False,
            )

        else:
            update_report_definition(
                workspace_id,
                item_id,
                item_definition=item_definition,
                df=False,
            )

    logger.success(
        f'All reports were deployed to workspace "{workspace}" successfully.'
    )
    return None


def convert_report_definition_to_by_path(
    report_path: Union[str, Path],
    workspace_path: Union[str, Path],
) -> None:
    """
    Convert report definition to use byPath reference.

    Args:
        report_path (Union[str, Path]): The file path to the report.
        workspace_path (Union[str, Path]): The file path to the workspace.
    """
    # Read the current definition.pbir
    definition_path = f'{report_path}/definition.pbir'

    if not os.path.exists(definition_path):
        logger.warning(f'definition.pbir not found: {definition_path}')
        return None

    try:
        with open(definition_path, 'r', encoding='utf-8') as f:
            report_definition = json.load(f)
    except Exception as e:
        logger.error(f'Error reading definition.pbir: {e}')
        return None

    # Check if it already uses byPath
    dataset_reference = report_definition.get('datasetReference', {})

    if 'byPath' in dataset_reference:
        logger.info(f'Report already uses byPath reference.')
        return None

    if 'byConnection' not in dataset_reference:
        logger.warning(f'Report has no byConnection reference.')
        return None

    # Extract semantic model name from connection string
    connection_string = dataset_reference['byConnection'].get(
        'connectionString', ''
    )

    # Capture the value after "initial catalog="
    match = re.search(r'initial catalog=([^;]+)', connection_string)
    if not match:
        logger.warning(
            f'Could not extract semantic model name from connection string.'
        )
        return None

    semantic_model_name = match.group(1).strip('"')

    logger.info(f'Found semantic model: {semantic_model_name}')

    # Find the semantic model directory relative to the report
    # Look for *.SemanticModel directories in the project
    semantic_model_pattern = (
        f'{workspace_path}/**/{semantic_model_name}.SemanticModel'
    )
    semantic_model_paths = glob.glob(semantic_model_pattern, recursive=True)

    if not semantic_model_paths:
        logger.error(
            f'Semantic model directory not found: {semantic_model_name}.SemanticModel'
        )
        return None

    if len(semantic_model_paths) > 1:
        logger.warning(
            f'Multiple semantic model directories found for {semantic_model_name}, using first one'
        )

    semantic_model_path = semantic_model_paths[0]
    logger.info(f'Found semantic model at: {semantic_model_path}')

    # Calculate relative path from definition.pbir to semantic model
    # The definition.pbir is inside the Report directory, so we need to go up one level first
    definition_dir = report_path  # This is the .Report directory
    relative_path = os.path.relpath(semantic_model_path, definition_dir)

    # Convert backslashes to forward slashes for consistency
    relative_path = relative_path.replace('\\', '/')

    logger.info(f'Relative path from definition.pbir: {relative_path}')

    # Update the dataset reference
    new_dataset_reference = {'byPath': {'path': relative_path}}

    # Create updated definition
    updated_definition = report_definition.copy()
    updated_definition['datasetReference'] = new_dataset_reference

    # Write the updated definition back to file
    try:
        with open(definition_path, 'w', encoding='utf-8') as f:
            json.dump(updated_definition, f, indent=2)

        logger.success(
            f'Successfully converted report to use byPath reference'
        )

    except Exception as e:
        logger.error(f'Error writing updated definition.pbir: {e}')

    return None


REPORT_DEFINITION = """{
  "$schema": "https://developer.microsoft.com/json-schemas/fabric/item/report/definitionProperties/2.0.0/schema.json",
  "version": "4.0",
  "datasetReference": {
    "byConnection": {
      "connectionString": "Data Source=powerbi://api.powerbi.com/v1.0/myorg/#{workspace_name}#;initial catalog=#{semantic_model_name}#;integrated security=ClaimsToken;semanticmodelid=#{semantic_model_id}#"
    }
  }
}"""


def convert_report_definition_to_by_connection(
    workspace_name: Union[str, Path],
    report_path: Union[str, Path],
) -> None:
    """
    Convert report definition to use byConnection reference.

    Args:
        workspace_name: The name of the Power BI workspace.
        report_path: The file path to the report.
    """
    with open(f'{report_path}/definition.pbir', 'r') as f:
        report_definition = json.load(f)

    dataset_reference = report_definition['datasetReference']

    if 'byPath' in dataset_reference:
        dataset_path = dataset_reference['byPath']['path']
        dataset_name = dataset_path.split('/')[-1].split('.SemanticModel')[0]

    elif 'byConnection' in dataset_reference:
        text_to_search = dataset_reference['byConnection']['connectionString']
        # Capture the value after "initial catalog="
        match = re.search(r'initial catalog=([^;]+)', text_to_search)
        if match:
            dataset_name = match.group(1)

    print(f'Semantic model: {dataset_name}')

    # Get the semantic model ID
    semantic_model_id = get_semantic_model_id(workspace_name, dataset_name)
    if semantic_model_id is None:
        logger.error(f'Could not find semantic model ID for {dataset_name}')
        return None

    report_definition_template = REPORT_DEFINITION

    report_definition_updated = report_definition_template.replace(
        '#{workspace_name}#', workspace_name
    )
    report_definition_updated = report_definition_updated.replace(
        '#{semantic_model_name}#', dataset_name
    )
    report_definition_updated = report_definition_updated.replace(
        '#{semantic_model_id}#', semantic_model_id
    )

    # Write the updated report definition to the definition.pbir
    with open(f'{report_path}/definition.pbir', 'w', encoding='utf-8') as f:
        f.write(report_definition_updated)



from ..api.api import api_request
from ..utils.exceptions import ResourceNotFoundError
from ..utils.logging import get_logger

logger = get_logger(__name__)


def get_user_id(email):
    """Get user ID by email"""
    response = api_request(
        endpoint=f'/users/{email}',
        audience='graph',
    )
    if response.get('id'):
        logger.success(f"User {email} found with ID: {response['id']}")
        return response['id']
    else:
        raise ResourceNotFoundError(f'User {email} not found')


def get_user_email(id):
    """Get user email by ID"""
    response = api_request(
        endpoint=f'/users/{id}',
        audience='graph',
    )
    if response.get('userPrincipalName'):
        logger.success(
            f"User {id} found with email: {response['userPrincipalName']}"
        )
        return response['userPrincipalName']
    else:
        raise ResourceNotFoundError(f'User {id} not found')


def get_security_group_id(group_name):
    """Get group ID by display name or mail"""
    response = api_request(
        endpoint=f"/groups?$filter=displayName eq '{group_name}'",
        audience='graph',
    )
    id = response['value'][0]['id']
    if id:
        logger.success(f'Group {group_name} found with ID: {id}')
        return id
    else:
        raise ResourceNotFoundError(f'Group {group_name} not found')


def get_security_group_name(group_id):
    """Get group display name by ID"""
    response = api_request(
        endpoint=f"/groups?$filter=id eq '{group_id}'",
        audience='graph',
    )
    name = response['value'][0]['displayName']
    if name:
        logger.success(f'Group {group_id} found with name: {name}')
        return name
    else:
        raise ResourceNotFoundError(f'Group {group_id} not found')


def get_app_registration_id(app_name):
    """Get application (app registration) ID by display name"""
    response = api_request(
        endpoint=f"/applications?$filter=displayName eq '{app_name}'",
        audience='graph',
    )
    id = response['value'][0]['id']
    if id:
        logger.success(f'App registration {app_name} found with ID: {id}')
        return id
    else:
        raise ResourceNotFoundError(f'App registration {app_name} not found')


def get_app_registration_name(app_id):
    """Get application (app registration) ID by display name"""
    response = api_request(
        endpoint=f"/applications?$filter=id eq '{app_id}'",
        audience='graph',
    )
    name = response['value'][0]['displayName']
    if name:
        logger.success(f'App registration {app_id} found with name: {name}')
        return name
    else:
        raise ResourceNotFoundError(f'App registration {app_id} not found')


def get_service_principal_id(spn_name):
    """Get application (app registration) ID by display name"""
    response = api_request(
        endpoint=f"/servicePrincipals?$filter=displayName eq '{spn_name}'",
        audience='graph',
    )
    id = response['value'][0]['id']
    if id:
        logger.success(f'Service Principal {spn_name} found with ID: {id}')
        return id
    else:
        raise ResourceNotFoundError(f'Service Principal {spn_name} not found')


def get_service_principal_name(spn_id):
    """Get application (app registration) ID by display name"""
    response = api_request(
        endpoint=f"/servicePrincipals?$filter=id eq '{spn_id}'",
        audience='graph',
    )
    name = response['value'][0]['displayName']
    if name:
        logger.success(f'Service Principal {spn_id} found with name: {name}')
        return name
    else:
        raise ResourceNotFoundError(f'Service Principal {spn_id} not found')



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import (
    create_folders_from_path_string,
    resolve_folder_from_id_to_path,
)
from ..items.environments import (
    create_environment,
    get_environment,
    get_environment_definition,
    list_environments,
    resolve_environment,
    update_environment_definition,
)
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import (
    delete_path,
    extract_display_name_from_platform,
    extract_middle_path,
    list_paths_of_type,
    pack_item_definition,
    unpack_item_definition,
)

logger = get_logger(__name__)


def get_environment_config(
    workspace: str, environment: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific environment config from a workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        environment (str): The name or ID of the semantic.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the environment
    """
    item = environment
    item_data = get_environment(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('displayName')] = {}

        config = {
            'id': item_data['id'],
            'description': item_data.get('description', None),
            'folder_id': ''
            if item_data.get('folderId') is None
            or pd.isna(item_data.get('folderId'))
            else item_data['folderId'],
        }

        return config


def get_all_environments_config(
    workspace: str,
) -> Union[Dict[str, Any], None]:
    """
    Get environments config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config of all environments in the workspace
    """
    items = list_environments(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:

        item_data = get_environment(workspace, item['id'], df=False)

        config[item['displayName']] = {
            'id': item['id'],
            'description': item.get('description', None),
            'folder_id': ''
            if item.get('folderId') is None or pd.isna(item.get('folderId'))
            else item['folderId'],
        }

    return config


def export_environment(
    workspace: str,
    environment: str,
    path: Union[str, Path],
) -> None:
    """
    Export a environment to path.

    Args:
        workspace (str): The name or ID of the workspace.
        environment (str): The name or ID of the environment.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    item = get_environment(workspace_id, environment, df=False)
    try:
        folder_path = resolve_folder_from_id_to_path(
            workspace_id, item['folderId']
        )
    except:
        logger.info(
            f'{item["displayName"]}.Environment is not inside a folder.'
        )
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + '.Environment')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + '.Environment')
        )
    os.makedirs(item_path, exist_ok=True)

    definition = get_environment_definition(workspace_id, item['id'])
    if definition is None:
        return None

    unpack_item_definition(definition, item_path)

    logger.success(
        f'`{item["displayName"]}.Environment` was exported to {item_path} successfully.'
    )
    return None


def export_all_environments(
    workspace: str,
    path: Union[str, Path],
) -> None:
    """
    Export a environment to path.

    Args:
        workspace (str): The name or ID of the workspace.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_environments(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, item['folderId']
            )
        except:
            logger.info(
                f'{item["displayName"]}.Environment is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + '.Environment')
        else:
            item_path = (
                Path(path)
                / folder_path
                / (item['displayName'] + '.Environment')
            )
        os.makedirs(item_path, exist_ok=True)

        definition = get_environment_definition(workspace_id, item['id'])
        if definition is None:
            return None

        unpack_item_definition(definition, item_path)

    logger.success(f'All environments were exported to {path} successfully.')
    return None


@df
def deploy_environment(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Deploy a environment to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the environment.
        start_path (Optional[str]): The starting path for folder creation.
        description (Optional[str]): Description for the environment.
        df (Optional[bool]): If True, returns a DataFrame, otherwise returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The deployed environment or None if deployment fails.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    display_name = extract_display_name_from_platform(path)
    if display_name is None:
        return None

    environment_id = resolve_environment(workspace_id, display_name)

    folder_path_string = extract_middle_path(path, start_path=start_path)
    folder_id = create_folders_from_path_string(
        workspace_id, folder_path_string
    )

    item_definition = pack_item_definition(path)

    if environment_id is None:
        return create_environment(
            workspace_id,
            display_name=display_name,
            item_definition=item_definition,
            description=description,
            folder=folder_id,
            df=False,
        )

    else:
        return update_environment_definition(
            workspace_id,
            environment_id,
            item_definition=item_definition,
            df=False,
        )


def deploy_all_environments(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all environments to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the environments.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    environments_paths = list_paths_of_type(path, 'Environment')

    for path_ in environments_paths:

        display_name = extract_display_name_from_platform(path_)
        if display_name is None:
            return None

        environment_id = resolve_environment(workspace_id, display_name)

        folder_path_string = extract_middle_path(path_, start_path=start_path)
        folder_id = create_folders_from_path_string(
            workspace_id, folder_path_string
        )

        item_definition = pack_item_definition(path_)

        if environment_id is None:
            create_environment(
                workspace_id,
                display_name=display_name,
                item_definition=item_definition,
                folder=folder_id,
                df=False,
            )

        else:
            update_environment_definition(
                workspace_id,
                environment_id,
                item_definition=item_definition,
                df=False,
            )

    logger.success(
        f'All environments were deployed to workspace "{workspace}" successfully.'
    )
    return None


def _create_environment_external_library_yaml(
    libraries: List[tuple[str, str]],
):

    env_yaml = """dependencies:
  - pip:"""

    for library, version in libraries:
        env_yaml += f"""
      - {library}=={version}"""

    target_path = '../tmp/env/Libraries/PublicLibraries/environment.yaml'

    os.makedirs(os.path.dirname(target_path), exist_ok=True)
    with open(target_path, 'w', encoding='utf-8') as f:
        f.write(env_yaml)
    logger.debug(f'Environment external libraries were created successfully.')
    return None


def add_environment_external_library_from_pypi(
    workspace: str,
    environment: str,
    libraries: List[tuple[str, str]],
):
    """
    Add external libraries to an environment from PyPI.

    Args:
        workspace (str): The name or ID of the workspace.
        environment (str): The name or ID of the environment.
        libraries (List[tuple[str, str]]): A list of tuples containing library names and their versions.

    Returns:
        None

    Examples:
    ```python
        add_environment_external_library_from_pypi(
            workspace='FabricOpsFlow-DEV',
            environment='Default',
            libraries=[
                ('pyFabricOps', '0.3.3'),
                ('pandas', '1.5.3'),
                ('numpy', '1.24.2')
            ]
        )
    ```
    """
    target_path = '../tmp/env'

    definition = get_environment_definition(
        workspace,
        environment,
    )

    delete_path('../tmp')

    unpack_item_definition(
        definition,
        path=target_path,
    )

    _create_environment_external_library_yaml(libraries)

    update_environment_definition(
        workspace=workspace,
        environment=environment,
        environment_definition=pack_item_definition(target_path),
    )

    logger.success(
        f'External libraries were added to environment {environment} successfully.'
    )
    return None



import os
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas
from pandas import DataFrame

from ..core.folders import create_folder, list_folders, resolve_folder
from ..core.workspaces import resolve_workspace
from ..utils.logging import get_logger

logger = get_logger(__name__)


def generate_folders_paths(
    folders_df: DataFrame,
) -> DataFrame:
    """
    Returns the full path for the folder `folder_id` recursively concatenating the names of its parents.

    Args:
        folders_df (DataFrame): The DataFrame containing folder information.

    Returns:
        DataFrame: The full folder paths.
    """

    df = folders_df

    # Create a dict to lookup: id → {displayName, parentFolderId}
    folder_map = df.set_index('id')[['displayName', 'parentFolderId']].to_dict(
        'index'
    )

    # Recursive function with cache to build the full path
    @lru_cache(maxsize=None)
    def _build_full_path(folder_id: str) -> str:
        """
        Returns the full path for the folder `folder_id`,
        recursively concatenating the names of its parents.
        """
        node = folder_map.get(folder_id)
        if node is None:
            return ''  # id not found
        name = node['displayName']
        parent = node['parentFolderId']
        # If without parent, is root
        if pandas.isna(parent) or parent == '':
            return name
        # Otherwise, joins the parent path with self name
        return _build_full_path(parent) + '/' + name

    # Apply the function by each dataframe row
    df['folder_path'] = df['id'].apply(lambda x: _build_full_path(x))

    df = df.rename(columns={'id': 'folder_id'})
    return df[['folder_id', 'folder_path']]


def get_folders_paths(workspace: str) -> DataFrame:
    """
    Get the full folder paths for all folders in the workspace.

    Args:
        workspace (str): The workspace name.

    Returns:
        DataFrame: A DataFrame with folder IDs and their full paths.
    """
    folders_df = list_folders(workspace)

    if folders_df is None or folders_df.empty:
        logger.debug(f'No folders found in workspace {workspace}.')
        return None

    if 'parentFolderId' not in folders_df.columns:
        folders_df['parentFolderId'] = ''

    return generate_folders_paths(folders_df)


def get_folders_config(workspace: str) -> Union[Dict[str, Any], None]:
    """
    Get the folder configuration for a specific workspace.

    Args:
        workspace (str): The workspace name or ID.

    Returns:
        (Union[Dict[str, Any], None]): The folder configuration or None if not found.
    """
    folders = get_folders_paths(workspace)
    if folders is None:
        return None

    return folders.to_dict(orient='records')


def export_folders(workspace: str, path: Union[str, Path]) -> None:
    """
    Export all folders from a workspace to a specified path
    """
    folders = get_folders_paths(workspace)
    folders_list = folders.to_dict(orient='records')
    for folder in folders_list:
        folder_path_ = Path(path) / folder['folder_path']
        os.makedirs(folder_path_, exist_ok=True)
        # Create a dummy README.md in each created folder
        with open(
            Path(folder_path_) / 'README.md', 'w', encoding='utf-8'
        ) as f:
            f.write(
                f'# {folder["folder_path"]}\n\nThis folder corresponds to the Fabric workspace folder: **{folder["folder_path"]}**\n'
            )
    logger.success(
        f'All folders from workspace {workspace} were exported to {path} successfully.'
    )


def resolve_folder_from_id_to_path(
    workspace: str, folder_id: str
) -> Union[str, None]:
    """
    Return the folder path to the folder_id given for a specified worspace.
    """
    folders = get_folders_paths(workspace)
    if folders is None:
        return None

    folder_path = folders[folders['folder_id'] == folder_id][
        'folder_path'
    ].iloc[0]

    if folder_path is None:
        logger.info(f'{folder_id} not found in the workspace {workspace}')
        return None

    return folder_path


def deploy_folders(
    workspace: str,
    path: Union[str, Path],
):
    """
    Creates folders in Fabric workspace based on local folder structure

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the project directory.
    """
    if not os.path.exists(path):
        logger.error(f'Path {path} does not exist.')
        return None

    # Resolve workspace ID
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    # Get all local folders that contain Fabric artifacts
    fabric_artifacts = [
        '.SemanticModel',
        '.Report',
        '.Dataflow',
        '.Lakehouse',
        '.Warehouse',
        '.Notebook',
        '.DataPipeline',
    ]

    def _has_fabric_artifacts(path):
        """Check if folder or any subfolder contains Fabric artifacts"""
        for root, dirs, files in os.walk(path):
            for dir_name in dirs:
                if any(
                    dir_name.endswith(artifact)
                    for artifact in fabric_artifacts
                ):
                    return True
        return False

    # First pass: identify folders with artifacts and their parent folders
    folders_with_artifacts = set()

    for root, dirs, files in os.walk(path):
        for dir_name in dirs:
            full_path = os.path.join(root, dir_name)

            # Check if this folder has Fabric artifacts
            if _has_fabric_artifacts(full_path):
                relative_path = os.path.relpath(full_path, path).replace(
                    '\\', '/'
                )
                folders_with_artifacts.add(relative_path)

                # Also mark all parent folders as needed
                parent_path = os.path.dirname(relative_path).replace('\\', '/')
                while (
                    parent_path != path
                    and parent_path != '.'
                    and parent_path != ''
                ):
                    folders_with_artifacts.add(parent_path)
                    parent_path = os.path.dirname(parent_path).replace(
                        '\\', '/'
                    )

    # Second pass: build folder list only for folders with artifacts
    local_folders = []
    for root, dirs, files in os.walk(path):
        for dir_name in dirs:
            full_path = os.path.join(root, dir_name)
            relative_path = os.path.relpath(full_path, path).replace('\\', '/')

            # Only include folders that contain artifacts or are parents of folders with artifacts
            if relative_path in folders_with_artifacts:
                # Calculate depth for proper ordering (parents before children)
                depth = relative_path.count('/')

                # Get parent folder name (not full path)
                parent_relative_path = os.path.dirname(relative_path).replace(
                    '\\', '/'
                )
                parent_folder_name = None
                if (
                    parent_relative_path
                    and parent_relative_path != '.'
                    and parent_relative_path != ''
                ):
                    parent_folder_name = os.path.basename(parent_relative_path)

                local_folders.append(
                    {
                        'path': relative_path,
                        'name': dir_name,
                        'full_path': full_path,
                        'depth': depth,
                        'parent_path': parent_relative_path,
                        'parent_name': parent_folder_name,
                    }
                )

    # Sort by depth to ensure parent folders are created first
    local_folders.sort(key=lambda x: x['depth'])

    logger.info(
        f'Found {len(local_folders)} folders containing Fabric artifacts'
    )

    # Keep track of created folders by path -> folder_id
    created_folders = {}

    for folder_info in local_folders:
        folder_name = folder_info['name']
        parent_path = folder_info['parent_path']
        parent_name = folder_info['parent_name']

        # Determine parent folder ID from previously created folders
        parent_folder_id = None
        if parent_path and parent_path in created_folders:
            parent_folder_id = created_folders[parent_path]

        # Create folder in Fabric
        if parent_folder_id:
            create_folder(
                workspace, folder_name, parent_folder=parent_folder_id
            )
        elif parent_name:
            create_folder(workspace, folder_name, parente_folder=parent_name)
        else:
            create_folder(workspace, folder_name)

    logger.success(f'Created all folders in the workspace {workspace}.')


def create_folders_from_path_string(workspace: str, path: str) -> str:
    """
    Create recursively folders and subfolders from a path string.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The name or ID of the folder.

    Returns:
        str: The ID of the final folder.
    """
    workspace_id = resolve_workspace(workspace)

    if path is None or '/' not in path:
        return None

    folders_tree = path.split('/')

    parent_folder_id = None

    for folder in folders_tree:

        # Get folder_id if folder exists
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id is not None:
            logger.info(
                f'Folder `{folder}` already exists with ID `{folder_id}`.'
            )

        # If not, creates it.
        else:
            folder_id = create_folder(
                workspace_id,
                folder,
                parent_folder=parent_folder_id,
                df=False,
            ).get('id')
            logger.success(
                f'Folder `{folder}` created with ID `{folder_id}` successfully.'
            )

        parent_folder_id = folder_id

    return folder_id



from typing import Any, Dict

from ..core.workspaces import get_workspace, list_workspace_role_assignments
from ..utils.exceptions import ResourceNotFoundError


def get_workspace_config(
    workspace: str,
) -> Dict[str, Any]:
    """
    Retrieves the workspace details for a given workspace.

    Args:
        workspace (str): The ID or name of the workspace to retrieve configuration for.

    Returns:
        dict: A dictionary containing the workspace details, including workspace ID, name, description, capacity ID, region, and roles.

    Examples:
        ```python
        get_workspace_details('123e4567-e89b-12d3-a456-426614174000')
        get_workspace_details('MyProject')
        ```
    """
    # Retrieving details from the workspace
    workspace_details = get_workspace(workspace, df=False)
    if not workspace_details:
        raise ResourceNotFoundError(f'Workspace {workspace} not found.')

    workspace_name = workspace_details.get('displayName', '')
    workspace_id = workspace_details.get('id', '')
    workspace_description = workspace_details.get('description', '')
    capacity_id = workspace_details.get('capacityId', '')
    capacity_region = workspace_details.get('capacityRegion', '')

    # Retrieving workspace roles
    # Retrieve details
    roles_details = list_workspace_role_assignments(workspace_id, df=False)

    # Init a empty list
    roles = []

    # Iterate for each role details
    for role in roles_details:
        principal_type = role['principal']['type']
        role_entry = {
            'user_uuid': role['id'],
            'user_type': principal_type,
            'role': role['role'],
            'display_name': role['principal'].get('displayName', ''),
        }

        if principal_type == 'Group':
            group_details = role['principal'].get('groupDetails', {})
            role_entry['group_type'] = group_details.get('groupType', '')
            role_entry['email'] = group_details.get('email', '')
        elif principal_type == 'User':
            user_details = role['principal'].get('userDetails', {})
            role_entry['user_principal_name'] = user_details.get(
                'userPrincipalName', ''
            )
        elif principal_type == 'ServicePrincipal':
            spn_details = role['principal'].get('servicePrincipalDetails', {})
            role_entry['app_id'] = spn_details.get('aadAppId', '')

        roles.append(role_entry)

    # Create a empty dict
    workspace_config = {}

    # Populate the dict
    workspace_config['workspace_id'] = workspace_id
    workspace_config['workspace_name'] = workspace_name
    workspace_config['workspace_description'] = workspace_description
    workspace_config['capacity_id'] = capacity_id
    workspace_config['capacity_region'] = capacity_region
    workspace_config['workspace_roles'] = roles

    return workspace_config



from typing import Any, Dict, List, Literal, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_tags(
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns a list of all the tenant's tags.

    Args:
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of tags in the workspace.

    Refs:
        https://learn.microsoft.com/en-us/rest/api/fabric/admin/tags/list-tags

    Examples:
        ```python
        list_tags()
        ```
    """
    resp = api_request(
        '/admin/tags',
        support_pagination=True,
    )
    return resp


def get_tag_id(tag_name: str) -> Union[str, None]:
    """
    Retrieves the ID of a tag by its name.

    Args:
        tag_name (str): The name of the tag.

    Returns:
        str | None: The ID of the tag if found, otherwise None.
    """
    tags = list_tags(df=False)
    for _tag in tags:
        if _tag['displayName'] == tag_name:
            return _tag['id']
    logger.warning(f"Tag '{tag_name}' not found.")
    return None


def resolve_tag(tag: str) -> Union[str, None]:
    """
    Resolves a tag name to its ID.

    Args:
        tag (str): The name or ID of the tag.

    Returns:
        str | None: The ID of the tag if found, otherwise None.
    """
    if is_valid_uuid(tag):
        return tag
    else:
        return get_tag_id(tag)


@df
def bulk_create_tags(
    payload: Dict[str, Any],
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Create a new domain in the tenant.

    Args:
        display_name (str): The name of the domain to create.
        description (str): The description of the domain to create.
        parent_domain (str): The name or ID of the parent domain.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created domain details if successful, otherwise None.

    Refs:
        https://learn.microsoft.com/en-us/rest/api/fabric/admin/tags/bulk-create-tags

    Examples:
        ```python

        # Create tags "Tag1" and "Tag2" under the "Tenant" scope
        bulk_create_tags(
            {
                "scope": {
                    "type": "Tenant"
                },
                "tags": [
                    {
                        "displayName": "Tag1"
                    },
                    {
                        "displayName": "Tag2"
                    }
                ]
            }
        )

        # Create tags "Tag1" and "Tag2" under the "Domain" scope
        bulk_create_tags(
            {
                "scope": {
                    "type": "Domain",
                    "domainId": "98f6b7c8-1234-5678-90ab-cdef12345678"
                },
                "tags": [
                    {
                        "displayName": "Tag1"
                    },
                    {
                        "displayName": "Tag2"
                    }
                ]
            }
        )
        ```
    """
    return api_request(
        '/admin/tags/bulkCreateTags',
        payload=payload,
        method='post',
    )


def delete_tag(tag: str) -> None:
    """
    Delete a tag by its name or ID.

    Args:
        tag (str): The name or ID of the tag to delete.

    Returns:
        None.

    Examples:
        ```python
        delete_tag('98f6b7c8-1234-5678-90ab-cdef12345678')
        ```

    Refs:
        https://learn.microsoft.com/en-us/rest/api/fabric/admin/tags/delete-tag
    """
    return api_request(
        '/admin/domains/' + resolve_tag(tag),
        method='delete',
    )


@df
def update_tag(
    tag: str,
    display_name: str = None,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the specified tag.

    Args:
        tag (str): The name or id of the current tag.
        display_name (str): The new name of the tag to update.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated tag details if successful, otherwise None.

    Examples:
        ```python
        update_tag(
            '98f6b7c8-1234-5678-90ab-cdef12345678',
            display_name='New Tag Name',
        )
        ```
    """
    payload = {'displayName': display_name}

    return api_request(
        '/admin/domains/' + resolve_tag(tag),
        payload=payload,
        method='patch',
    )
